# -*- coding: utf-8 -*-
"""
Tools for calculating the shock response spectrum. Adapted and
enhanced from the Yeti version.
"""

import itertools as it
import multiprocessing as mp
import ctypes
import os
from math import sin, cos, exp, sqrt, pi
from warnings import warn
import numpy as np
import scipy.signal as signal
import scipy.interpolate as interp
from pyyeti import dsp, psd


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


#   SRS types:
#    - absolute acceleration
#    - relative acceleration
#    - relative velocity
#    - pseudo velocity
#    - relative displacement
#
#   Recursive relation:
#    yn = b0 xn  +  b1 xn-1  + b2 xn-2  -  a1 yn-1  -  a2 yn-2
#
#     a is always the same
#
#  Steps to derive coefficients follow (note that the coefficients for
#  wn != 0 were taken from reference 1). The following is here for
#  reference.
#
#   1. Write transfer function H(s) for item of interest (minus sign
#      because right-hand-side is -base_acceleration):
#         absacce = -(c s + k) / (s**2 + c s + k)
#         reldisp = -1 / (s**2 + c s + k)
#         relvelo = -s / (s**2 + c s + k)
#         relacce = -s**2 / (s**2 + c s + k)
#
#   2. For:
#        Impulse Invariant Coefs: compute inverse Laplace of H(s)
#        Step Invariant Coefs:    compute inverse Laplace of H(s)/s
#        Ramp Invariant Coefs:    compute inverse Laplace of H(s)/s**2
#
#   3. Convert t --> n dT
#
#   4. Compute z-transform
#
#   5. Multiply by:
#        Impulse Invariant Coefs: dT
#        Step Invariant Coefs:    (1-z**-1)
#        Ramp Invariant Coefs:    (1-z**-1)**2 / (dT*z**-1)
#
#   6. Simplify and extract coefficients
#
#  Note that for the step and ramp invariants, the transfer function
#  H(s) is integrated once or twice, and operated on like the impulse
#  invariant approach. Then, at the end, the result is modified by
#  dividing by the effect of the integration noting that:
#        1/s    --> 1/(1-z**-1)
#        1/s**2 --> dT*z**-1 / (1-z**-1)**2
#
#  For example, the ramp invariant reldisp coefficients for wn == 0
#  were computed as outlined above:
#   1. H(s) = -1/s**2
#   2. inv-Laplace of H(s)/s**2 = invL[-1/s**4] = -t**3 / 6
#   3. Z[t**3 / 6] = -dT**3 / 6 * z**-1 * (1 + 4*z**-1 + z**-2) /
#                          (1-z**-1)**4
#   4. Mult by (1-z**-1)**2 / (dT*z**-1) and simplify:
#        H(z) = -dT**2 / 6 * (1 + 4*z**-1 + z**-2) /
#                            (1 - 2*z**-1 + z**-2)
#
#    References
#    ----------
#    .. [1] “Mechanical vibration and shock – Signal processing – Part
#           4: Shock-response spectrum analysis”, ISO 18431-4.
#
#    .. [2] Morin, A. and Labbé, P., "Derivation of Recursive Digital
#           Filters by the Step-Invariant and the Ramp-Invariant
#           Transformations", DREV R-4325/84, May 1984, UNCLASSIFIED.
#
#    .. [3] David Smallwood, "An Improved Recursive Formula for
#           Calculating Shock Response Spectra", 51st Shock and
#           Vibration Bulletin (1980).
#
#    .. [4] Kjell Ahlin, "Shock Response Spectrum Calculation - An
#           Improvement of the Smallwood Algorithm",
#           http://www.vibrationdata.com/tutorials/Ahlin_SRS.pdf

HIST_ = None
ICVALS_ = None
SIG_ = None
SRSmax_ = None
WN_ = None


def createSharedArray(dimensions, ctype=ctypes.c_double):
    """
    Creates array in shared memory segment and fills with zeros
    """
    shared_arr = mp.RawArray(ctype, int(np.prod(dimensions)))
    return shared_arr


def copyToSharedArray(arr, ctype=ctypes.c_double):
    """
    Create array in shared memory segment
    """
    shared_arr = mp.RawArray(ctype, arr.size)
    # convert to numpy array (shared memory) and reshape:
    a = np.frombuffer(shared_arr).reshape(arr.shape)
    a[:] = arr
    return shared_arr


def absacce(Q, dT, wn):
    """
    Utility routine used by :func:`srs` to get absolute acceleration
    digital filter coefficients. Returns (b, a) for use in
    :func:`scipy.signal.lfilter`.
    """
    zeta = 1 / 2 / Q
    sqz = sqrt(1 - zeta * zeta)
    wd = wn * sqz
    E = exp(-zeta * wn * dT)
    E2 = E * E
    B = dT * wd
    C = E * cos(B)
    if wn == 0:
        b = np.array([0.0, 0.0, 0.0])
    else:
        S = E * sin(B)
        Sb = S / B
        beta0 = 1 - Sb
        beta1 = 2 * (Sb - C)
        beta2 = E2 - Sb
        b = np.array([beta0, beta1, beta2])
    a = np.array([1, -2 * C, E2])
    return b, a


def relacce(Q, dT, wn):
    """
    Utility routine used by :func:`srs` to get relative acceleration
    digital filter coefficients. Returns (b, a) for use in
    :func:`scipy.signal.lfilter`.
    """
    zeta = 1 / 2 / Q
    sqz = sqrt(1 - zeta * zeta)
    wd = wn * sqz
    E = exp(-zeta * wn * dT)
    E2 = E * E
    B = dT * wd
    C = E * cos(B)
    b = np.array([-1.0, 2.0, -1.0])
    if wn != 0.0:
        b *= (E * sin(B)) / B
    a = np.array([1, -2 * C, E2])
    return b, a


def reldisp(Q, dT, wn):
    """
    Utility routine used by :func:`srs` to get relative displacement
    digital filter coefficients. Returns (b, a) for use in
    :func:`scipy.signal.lfilter`.
    """
    zeta = 1 / 2 / Q
    E = exp(-zeta * wn * dT)
    E2 = E * E
    sqz = sqrt(1 - zeta * zeta)
    wd = wn * sqz
    B = dT * wd
    C = E * cos(B)
    if wn == 0:
        # See notes above for the derivation of these coefficients:
        b = np.array([-1.0, -4.0, -1.0]) * dT ** 2 / 6
    else:
        S = E * sin(B)
        f = dT * wn * wn * wn
        q = (2 * zeta * zeta - 1) / sqz
        beta0 = ((1 - C) / Q - q * S - wn * dT) / f
        beta1 = (2 * C * wn * dT - (1 - E2) / Q + 2 * q * S) / f
        beta2 = (-E2 * (wn * dT + 1 / Q) + C / Q - q * S) / f
        b = np.array([beta0, beta1, beta2])
    a = np.array([1, -2 * C, E2])
    return b, a


def pvelo(Q, dT, wn):
    """
    Utility routine used by :func:`srs` to get pseudo-velocity
    (relative displacement * omega) digital filter coefficients.
    Returns (b, a) for use in :func:`scipy.signal.lfilter`.
    """
    zeta = 1 / 2 / Q
    sqz = sqrt(1 - zeta * zeta)
    wd = wn * sqz
    E = exp(-zeta * wn * dT)
    E2 = E * E
    B = dT * wd
    C = E * cos(B)
    if wn == 0:
        b = np.array([0.0, 0.0, 0.0])
    else:
        S = E * sin(B)
        f = dT * wn * wn
        q = (2 * zeta * zeta - 1) / sqz
        beta0 = ((1 - C) / Q - q * S - wn * dT) / f
        beta1 = (2 * C * wn * dT - (1 - E2) / Q + 2 * q * S) / f
        beta2 = (-E2 * (wn * dT + 1 / Q) + C / Q - q * S) / f
        b = np.array([beta0, beta1, beta2])
    a = np.array([1, -2 * C, E2])
    return b, a


def pacce(Q, dT, wn):
    """
    Utility routine used by :func:`srs` to get pseudo-acceleration
    (relative displacement * omega^2) digital filter coefficients.
    Returns (b, a) for use in :func:`scipy.signal.lfilter`.
    """
    zeta = 1 / 2 / Q
    sqz = sqrt(1 - zeta * zeta)
    wd = wn * sqz
    E = exp(-zeta * wn * dT)
    E2 = E * E
    B = dT * wd
    C = E * cos(B)
    if wn == 0:
        b = np.array([0.0, 0.0, 0.0])
    else:
        S = E * sin(B)
        f = dT * wn
        q = (2 * zeta * zeta - 1) / sqz
        beta0 = ((1 - C) / Q - q * S - wn * dT) / f
        beta1 = (2 * C * wn * dT - (1 - E2) / Q + 2 * q * S) / f
        beta2 = (-E2 * (wn * dT + 1 / Q) + C / Q - q * S) / f
        b = np.array([beta0, beta1, beta2])
    a = np.array([1, -2 * C, E2])
    return b, a


def relvelo(Q, dT, wn):
    """
    Utility routine used by :func:`srs` to get relative velocity
    digital filter coefficients. Returns (b, a) for use in
    :func:`scipy.signal.lfilter`.
    """
    if wn == 0.0:
        b = np.array([-1.0, -1.0]) * dT / 2
        a = np.array([1.0, -1.0])
    else:
        zeta = 1 / 2 / Q
        sqz = sqrt(1 - zeta * zeta)
        wd = wn * sqz
        E = exp(-zeta * wn * dT)
        E2 = E * E
        B = dT * wd
        C = E * cos(B)
        S = E * sin(B)
        Sz = S * zeta / sqz
        f = dT * wn * wn
        beta0 = (C + Sz - 1) / f
        beta1 = (1 - E2 - 2 * Sz) / f
        beta2 = (E2 + Sz - C) / f
        a = np.array([1, -2 * C, E2])
        b = np.array([beta0, beta1, beta2])
    return b, a


def _absmeth(resp):
    return abs(resp).max(axis=0)


def _posmeth(resp):
    return abs(resp.max(axis=0))


def _possmeth(resp):
    return resp.max(axis=0)


def _negmeth(resp):
    return abs(resp.min(axis=0))


def _negsmeth(resp):
    return resp.min(axis=0)


def _rmsmeth(resp):
    return np.sqrt((resp ** 2).mean(axis=0))


def fftroll(sig, sr, ppc, frq):
    """
    Increase sample rate using FFT for :func:`srs`.

    Parameters
    ----------
    sig : ndarray
        The signal(s), time-steps x n.
    sr : scalar
        Sample rate.
    ppc : scalar
        Minimum points per cycle.
    frq : scalar
        Highest frequency of the SDOF system.

    Returns
    -------
    signew : ndarray
        The resampled version of `sig` that meets the `ppc`
        requirement.
    srnew : scalar
        The new sample rate.

    Notes
    -----
    Because of the very poor performance of the SciPy FFT for signals
    of unfortunate lengths, if the number of time-steps is odd, the
    last point is truncated. This should be only temporary and
    hopefully not too harmful.

    See also
    --------
    :func:`scipy.signal.resample`
    """
    N = sig.shape[0]
    if N > 1:
        curppc = sr / frq
        factor = int(np.ceil(ppc / curppc))
        if N & 1:
            sig = signal.resample(sig[:-1], factor * (N - 1), axis=0)
        else:
            sig = signal.resample(sig, factor * N, axis=0)
        sr *= factor
    return sig, sr


def lanroll(sig, sr, ppc, frq):
    """
    Increase sample rate using :func:`pyyeti.dsp.resample` for the SRS
    routine.

    Parameters
    ----------
    sig : ndarray
        The signal(s), time-steps x n.
    sr : scalar
        Sample rate.
    ppc : scalar
        Minimum points per cycle.
    frq : scalar
        Highest frequency of the SDOF system.

    Returns
    -------
    signew : ndarray
        The resampled version of `sig` that meets the ppc requirement.
    srnew : scalar
        The new sample rate.

    Notes
    -----
    The `pts` parameter for the :func:`pyyeti.dsp.resample` is set to
    65.  This was determined from trial and error and comparison to
    the FFT method.
    """
    N = sig.shape[0]
    if N > 1:
        curppc = sr / frq
        factor = int(np.ceil(ppc / curppc))
        sig = dsp.resample(sig, factor, 1, pts=65, axis=0)
        sr *= factor
    return sig, sr


def preroll(sig, sr, ppc, frq):
    """
    Apply pre-filter to account for attenuation due to insufficient
    sample rate.

    Parameters
    ----------
    sig : ndarray
        The signal(s), time-steps x n.
    sr : scalar
        Sample rate.
    ppc : scalar
        Minimum points per cycle.
    frq : scalar
        Highest frequency of the SDOF system.

    Returns
    -------
    signew : ndarray
        The filtered version of `sig`.
    srnew : scalar
        The new sample rate (unchanged from input).

    Notes
    -----
    The approach is scale the time-domain signal such that the
    roll-off is compensated for [#srs5]_.

    References
    ----------
    .. [#srs5] Kjell Ahlin, "Shock Response Spectrum Calculation - An
            Improvement of the Smallwood Algorithm",
            http://www.vibrationdata.com/tutorials/Ahlin_SRS.pdf
    """
    b = np.array([0.8767, 1.7533, 0.8767])
    a = np.array([1, 1.6296, 0.8111, 0.0659])
    sig = signal.filtfilt(b, a, sig, axis=0)
    return sig, sr


def linroll(sig, sr, ppc, frq):
    """
    Increase sample rate using linear interpolation for :func:`srs`.

    Parameters
    ----------
    sig : ndarray
        The signal(s), time-steps x n.
    sr : scalar
        Sample rate.
    ppc : scalar
        Minimum points per cycle.
    frq : scalar
        Highest frequency of the SDOF system.

    Returns
    -------
    signew : ndarray
        The resampled version of `sig` that meets the ppc requirement.
    srnew : scalar
        The new sample rate.

    Notes
    -----
    Note that linear interpolation is not recommended for increasing
    sample rate.
    """
    N = sig.shape[0]
    if N > 1:
        curppc = sr / frq
        factor = int(np.ceil(ppc / curppc))
        told = np.arange(N) / sr
        sr *= factor
        tnew = np.linspace(0.0, told[-1], N * factor - 1)
        ifunc = interp.interp1d(told, sig, axis=0)
        sig = ifunc(tnew)
    return sig, sr


def _mk_par_globals(wn, sig, srsmax, hist):
    global WN_, SIG_, SRSmax_, HIST_
    WN_ = np.frombuffer(wn[0]).reshape(wn[1])
    SIG_ = np.frombuffer(sig[0]).reshape(sig[1])
    SRSmax_ = np.frombuffer(srsmax[0]).reshape(srsmax[1])
    if hist[0] is not None:
        HIST_ = np.frombuffer(hist[0]).reshape(hist[1])


def _dosrs_nohist(args):
    """Utility routine for parallel processing for when
    `getresp` is False"""
    (j, (coeffunc, Q, dT, methfunc, S)) = args
    b, a = coeffunc(Q, dT, WN_[j])
    resphist = signal.lfilter(b, a, SIG_, axis=0)
    SRSmax_[j] = methfunc(resphist[S:])


def _dosrs(args):
    """Utility routine for parallel processing for when
    `getresp` is True"""
    (j, (coeffunc, Q, dT, methfunc, S)) = args
    b, a = coeffunc(Q, dT, WN_[j])
    resphist = signal.lfilter(b, a, SIG_, axis=0)
    SRSmax_[j] = methfunc(resphist[S:])
    HIST_[:, :, j] = resphist[S:]


def _mk_par_globals_ic(wn, sig, icvals, srsmax, hist):
    global WN_, SIG_, ICVALS_, SRSmax_, HIST_
    WN_ = np.frombuffer(wn[0]).reshape(wn[1])
    SIG_ = np.frombuffer(sig[0]).reshape(sig[1])
    ICVALS_ = np.frombuffer(icvals[0]).reshape(icvals[1])
    SRSmax_ = np.frombuffer(srsmax[0]).reshape(srsmax[1])
    if hist[0] is not None:
        HIST_ = np.frombuffer(hist[0]).reshape(hist[1])


def _dosrs_nohist_ic(args):
    """Utility routine for parallel processing for when
    `getresp` is False"""
    (j, (coeffunc, Q, dT, methfunc, S, stype)) = args
    b, a = coeffunc(Q, dT, WN_[j])
    resphist = signal.lfilter(b, a, SIG_, axis=0)
    if stype == "reldisp":
        resphist += ICVALS_ / WN_[j] ** 2
    elif stype == "pvelo":
        resphist += ICVALS_ / WN_[j]
    else:
        # stype == 'pacce' or 'absacce'
        resphist += ICVALS_
    SRSmax_[j] = methfunc(resphist[S:])


def _dosrs_ic(args):
    """Utility routine for parallel processing for when
    `getresp` is True"""
    (j, (coeffunc, Q, dT, methfunc, S, stype)) = args
    b, a = coeffunc(Q, dT, WN_[j])
    resphist = signal.lfilter(b, a, SIG_, axis=0)
    if stype == "reldisp":
        resphist += ICVALS_ / WN_[j] ** 2
    elif stype == "pvelo":
        resphist += ICVALS_ / WN_[j]
    else:
        # stype == 'pacce' or 'absacce'
        resphist += ICVALS_
    SRSmax_[j] = methfunc(resphist[S:])
    HIST_[:, :, j] = resphist[S:]


def _process_inputs(stype, peak, rolloff, time):
    """Utility routine for srs"""
    coefs = {
        "absacce": absacce,
        "relacce": relacce,
        "reldisp": reldisp,
        "relvelo": relvelo,
        "pvelo": pvelo,
        "pacce": pacce,
    }
    meth = {
        "pos": _posmeth,
        "neg": _negmeth,
        "abs": _absmeth,
        "rms": _rmsmeth,
        "poss": _possmeth,
        "negs": _negsmeth,
    }
    roll = {
        "fft": fftroll,
        "lanczos": lanroll,
        "prefilter": preroll,
        "linear": linroll,
        "none": None,
    }
    ptr = {"primary": 0, "total": 1, "residual": 2}
    coeffunc = coefs[stype]
    if isinstance(peak, str):
        methfunc = meth[peak]
    else:
        methfunc = peak
    if isinstance(rolloff, str):
        rollfunc = roll[rolloff]
    else:
        rollfunc = rolloff
    ptr = ptr[time]
    return coeffunc, methfunc, rollfunc, ptr


def _process_parallel(parallel, LF, size, maxcpu, getresp):
    """Utility routine for srs"""
    if parallel not in ["auto", "yes", "no"]:
        raise ValueError("invalid parallel option")
    if parallel != "no":
        ncpu = mp.cpu_count()
    if parallel == "auto":
        if (
            LF > 1
            and size > 50000
            and not getresp
            and ncpu > 1
            and not os.sys.platform.startswith("win")
        ):
            parallel = "yes"
        else:
            parallel = "no"
    if parallel == "yes":
        if maxcpu and ncpu > maxcpu:
            ncpu = maxcpu
        elif ncpu > 4:
            ncpu = (ncpu * 4) // 5
    else:
        ncpu = 1
    return parallel, ncpu


def _process_ic(sig, ic, stype):
    """Utility routine for srs"""
    doic = 0
    icvals = None
    s1 = sig[0]
    if ic == "shift":
        sig = sig - s1
    elif ic == "mshift":
        sig = sig - sig.mean(axis=0)
    elif ic == "steady":
        sig = sig - s1
        if stype == "absacce":
            icvals = s1
            doic = 1
        elif stype == "relacce" or stype == "relvelo":
            pass
        else:
            # 'reldisp', 'pvelo' or 'pacce'
            icvals = -s1
            doic = 1
    return sig, s1, doic, icvals


def _add_one_cycle(sig, freq, sr, H, ic, s1):
    """Utility routine for srs"""
    # append zeros to allow for one cycle of lowest non-zero
    # frequency
    pv = (freq > 0).nonzero()[0]
    if pv.size > 0:
        minf = freq[pv].min()
        nzeros = int(np.ceil(sr / minf))
        z = np.zeros((nzeros, H))
        if ic == "steady":
            sig = np.vstack((sig, z - s1))
        else:
            sig = np.vstack((sig, z))
    return sig, sig.shape[0]


def srs(
    sig,
    sr,
    freq,
    Q,
    ic="zero",
    stype="absacce",
    peak="abs",
    ppc=12,
    rolloff="lanczos",
    eqsine=False,
    time="primary",
    getresp=False,
    parallel="auto",
    maxcpu=14,
):
    r"""
    Shock response spectrum - response of single DOF systems to base
    excitation(s).

    Parameters
    ----------
    sig : 1d or 2d array_like
        Base acceleration signal; vector or matrix where column is a
        signal. If size is 1 x n (2d), that means there are n signals,
        each with length 1 (only initial conditions are
        calculated). For length 1 signal(s), you'll probably want to
        set `ic` to 'steady'.
    sr : scalar or None
        Sample rate. Can be None if signal(s) are length 1 and `ic` is
        not 'zero' (see note below).
    freq : 1d array_like
        Frequency vector in Hz. This defines the single DOF systems
        to use.
    Q : scalar > 0.5
        Dynamic amplification factor :math:`Q = 1/(2\zeta)` where
        :math:`\zeta` is the fraction of critical damping.
    ic : string; optional
        Specifies how to handle the initial conditions:

           ========   ===============================================
            `ic`      Initial conditions
           ========   ===============================================
           'zero'     uses zero initial conditions
           'shift'    shifts each signal to start at zero so there
                      are no step inputs and then uses zero initial
                      conditions
           'mshift'   shifts each signal by its mean, then uses zero
                      initial conditions
           'steady'   uses steady-state initial conditions
           ========   ===============================================

    stype : string; optional
        Specifies the type of response to recover:

           =========    =======================================
            `stype`     Response that :func:`srs` calculates
           =========    =======================================
           'absacce'    absolute acceleration
           'relacce'    relative acceleration
           'reldisp'    relative displacement
           'relvelo'    relative velocity
           'pvelo'      pseudo velocity (reldisp * omega)
           'pacce'      pseudo acceleration (reldisp * omega^2)
           =========    =======================================

    peak : string or function; optional
        If a string, it must be one of these values:

           ======    =============================
           `peak`    Value the :func:`srs` returns
           ======    =============================
           'abs'     absolute maximum
           'pos'     maximum, absolute value
           'neg'     minimum, absolute value
           'poss'    maximum, keeping signs
           'negs'    minimum, keeping signs
           'rms'     root-sum-square
           ======    =============================

        If a function, it must accept the 2d response ndarray with
        shape = (len(time), nsignals) and return a 1d array of "peaks"
        with shape = (nsignals,).
        For example, the 'abs' function is::

            def _absmeth(resp):
                return abs(resp).max(axis=0)

    ppc : scalar; optional
        Specifies the minimum points per cycle for SRS calculations.
        See also `rolloff`.

           ======    ==================================
           `ppc`     Maximum error at highest frequency
           ======    ==================================
               3     81.61%
               4     48.23%
               5     31.58%
              10     8.14% (minimum recommended `ppc`)
              12     5.67%
              15     3.64%
              20     2.05%
              25     1.31%
              50     0.33%
           ======    ==================================

    rolloff : string or function or None; optional
        Indicate which method to use to account for the SRS roll off
        when the minimum `ppc` value is not met. Either 'fft' or
        'lanczos' seem the best.  If a string, it must be one of these
        values:

           ===========    ==========================================
           `rolloff`      Notes
           ===========    ==========================================
           'fft'          Use FFT to upsample data as needed.  See
                          :func:`scipy.signal.resample`.
           'lanczos'      Use Lanczos resampling to upsample as
                          needed. See :func:`pyyeti.dsp.resample`.
           'prefilter'    Apply a high freq. gain filter to account
                          for the SRS roll-off. See
                          :func:`preroll` for more information. This
                          option ignores `ppc` [#srs4]_.
           'linear'       Use linear interpolation to increase the
                          points per cycle (this is not recommended;
                          it's only here as a test case).
           'none'         Don't do anything to enforce the minimum
                          `ppc`. Note error bounds listed above.
            None          Same as 'none'.
           ===========    ==========================================

        If a function, the call signature is:
        ``sig_new, sr_new = rollfunc(sig, sr, ppc, frq)``. `sig` is
        ``time x n``. The last three inputs are scalars. For example,
        the 'fft' function is (trimmed of documentation)::

            def fftroll(sig, sr, ppc, frq):
                N = sig.shape[0]
                if N > 1:
                    curppc = sr/frq
                    factor = int(np.ceil(ppc/curppc))
                    sig = signal.resample(sig, factor*N, axis=0)
                    sr *= factor
                return sig, sr

    eqsine : bool; optional
        If true, resulting history is divided by Q before the peak is
        extracted.
    time : string; optional
        Specifies the time-frame for SRS calculation:

           ==========   ============================================
            `time`      Time-frame for SRS calculation
           ==========   ============================================
           'primary'    During the signal(s) as input.
           'residual'   After the signal(s) (zeros are appended to
                        allow one cycle of the lowest frequency).
           'total'      During and after the signal(s) (primary &
                        residual).
           ==========   ============================================

    getresp : bool; optional
        If True, return the response time histories (see `resp` output
        below).
    parallel : string; optional
        Controls the parallelization of the SRS calculations:

           ==========   ============================================
           `parallel`   Notes
           ==========   ============================================
           'auto'       Routine determines whether or not to run
                        parallel.
           'no'         Do not use parallel processing.
           'yes'        Use parallel processing. Beware, depending
                        on the particular problem, using parallel
                        processing can be slower than not using it
                        (especially if `getresp` is True). On
                        Windows, be sure the :func:`srs` call is
                        contained within:
                        ``if __name__ == "__main__":``
           ==========   ============================================

    maxcpu : integer or None; optional
        Specifies maximum number of CPUs to use. If None, it is
        internally set to 4/5 of available CPUs (as determined from
        :func:`multiprocessing.cpu_count`.

    Returns
    -------
    sh : 1d or 2d ndarray
        The SRS results; ``sh.shape = (len(freq), nsignals)``. If
        `sig` is 1d, `sh` will also be 1d:
        ``sh.shape = (len(freq),)``.
    resp : dictionary; optional
        Only returned if `getresp` is True. Members:

        ======   =====================================================
         key     value
        ======   =====================================================
        't'      time vector for responses
        'sr'     sample rate associated with 't' (>= the input `sr`;
                 depends on inputs `sr`, `ppc`, and `rolloff`)
        'hist'   3-D array; shape = ``(len(t), nsignals, len(freq))``
        ======   =====================================================

    Notes
    -----
    The shock response spectrum is the response of single DOF
    system(s) that are excited by an input base acceleration::

                      _____    ^
                     |     |   |
                     |  M  |  ---  SDOF response (x)
                     |_____|
                      /  |
                    K \ |_| C  ^
                      /  |     |
                    [=======] ---  input base acceleration (sig)

    Derivation of the equation of motion follows. First, let:

    .. math::
        \begin{aligned}
        \ddot{z} &= sig \\
        M &= 1 \\
        K &= \omega^2 \\
        C &= 2\zeta\omega \\
        \end{aligned}

    Note that :math:`\omega=2 \pi f` where :math:`f` is a frequency in
    Hz from the input `freq`. The equation of motion is:

    .. math::
        \begin{aligned}
        \ddot{x} &= \sum Forces\; on\; M \\
        &= \omega^2(z-x)+2\zeta\omega(\dot{z}-\dot{x})
        \end{aligned}

    Define a relative coordinate :math:`u = x - z`. Then:

    .. math::
        \begin{aligned}
        \ddot{x}+2\zeta\omega\dot{u}+\omega^2 u &= 0 \\
        \ddot{u}+2\zeta\omega\dot{u}+\omega^2 u &= -\ddot{z}
        \end{aligned}

    In general, that equation is solved for each frequency for each
    signal, giving the relative displacement, velocity and
    acceleration.  The absolute acceleration is then calculated from:

    .. math::
        \ddot{x} = \ddot{u} + \ddot{z}

    By assuming the input signal is linear between time points (ramp
    invariant), these equations can be solved in closed form.
    Reference [#srs1]_ below has done this and conveniently put the
    solution in terms of linear digital filter coefficients.
    Furthermore, coefficients are provided to solve for whichever
    response is requested. More information on coefficient derivation
    can be found in references [#srs2]_, and [#srs3]_.  Reference
    [#srs4]_ is a method for accounting for rolloff.

    The maximum errors listed above are a summation of the bias error
    from the ramp invariant solver and the maximum error that can
    occur when selecting peaks (since peaks occur between solution
    points).  The error equations are (noting that
    :math:`sr/f = ppc`):

    .. math::
        \begin{aligned}
        &\text{bias error} = 1 - \left[ \sin \left( \frac{\pi f}{sr}
        \right) / \frac{\pi f}{sr} \right]^2 \\
        &\text{max peak error} = 1 - \cos \left(
        \frac{\pi f}{sr} \right)
        \end{aligned}

    Or:

    .. math::
        \begin{aligned}
        &\text{bias error} = 1 - \left[ \sin \left( \frac{\pi}{ppc}
        \right) / \frac{\pi}{ppc} \right]^2 \\
        &\text{max peak error} = 1 - \cos \left(
        \frac{\pi}{ppc} \right)
        \end{aligned}

    .. note::
        The 'zero' and 'mshift' initial conditions may be handled in a
        slightly different manner than one might think: the zero
        initial displacement and velocity conditions occur one time
        step before `sig` begins (where `sig` is also assumed
        zero). This means one time step is analyzed even if the
        signal(s) are length 1; this is why `sr` cannot be None when
        ``ic = 'zero'`` ('mshift' is okay because it behaves like
        'shift' when there is only one time step).

    .. note::
        In addition to the example shown below, this routine is
        demonstrated in the pyYeti :ref:`tutorial`:
        :doc:`/tutorials/srs`. There is also a link to the source
        Jupyter notebook at the top of the tutorial.

    References
    ----------
    .. [#srs1] “Mechanical vibration and shock – Signal processing –
           Part 4: Shock-response spectrum analysis”, ISO 18431-4.

    .. [#srs2] Morin, A. and Labbé, P., "Derivation of Recursive
           Digital Filters by the Step-Invariant and the Ramp-
           Invariant Transformations", DREV R-4325/84, May 1984,
           UNCLASSIFIED.

    .. [#srs3] David Smallwood, "An Improved Recursive Formula for
           Calculating Shock Response Spectra", 51st Shock and
           Vibration Bulletin (1980).

    .. [#srs4] Kjell Ahlin, "Shock Response Spectrum Calculation - An
           Improvement of the Smallwood Algorithm",
           http://www.vibrationdata.com/tutorials/Ahlin_SRS.pdf

    Raises
    ------
    ValueError
        When `sr` is None but number of time steps is greater than 1
        OR `ic` is set to 'zero'.

    Examples
    --------
    .. plot::
        :context: close-figs

        >>> from pyyeti import srs
        >>> import numpy as np
        >>> sr = 1000.
        >>> t = np.arange(0, 5, 1/sr)
        >>> sig = np.sin(2*np.pi*15*t)
        >>> Q = 20
        >>> frq = [10, 15, 20]
        >>> sh = srs.srs(sig, sr, frq, Q)
        >>> print(f'{sh[1]:.1f}')
        20.0

        Compare the upsampling/rolloff methods:

        >>> import matplotlib.pyplot as plt
        >>> _ = plt.figure('Example')
        >>> plt.clf()
        >>> sr = 200
        >>> t = np.arange(0, 5, 1/sr)
        >>> sig = np.sin(2*np.pi*15*t) + 3*np.sin(2*np.pi*85*t)
        >>> Q = 50
        >>> frq = np.linspace(5, 100, 476)
        >>> for meth in ['none', 'linear', 'prefilter',
        ...              'lanczos', 'fft']:
        ...    sh = srs.srs(sig, sr, frq, Q, rolloff=meth)
        ...    _ = plt.plot(frq, sh, label=meth)
        >>> _ = plt.legend(loc='best')
        >>> ttl = '85 Hz peak should approach 150'
        >>> _ = plt.title(ttl)
        >>> _ = plt.grid(True)
    """
    (coeffunc, methfunc, rollfunc, ptr) = _process_inputs(stype, peak, rolloff, time)
    freq = np.atleast_1d(freq)
    wn = 2 * pi * freq
    LF = len(freq)
    sig = np.atleast_1d(sig)
    if sig.ndim == 1:
        oneD = True
        sig = sig.reshape(-1, 1)
    else:
        oneD = False
    N = sig.shape[0]  # number of time steps
    H = sig.shape[1]  # number of histories

    if sr is None:
        if N > 1 or ic == "zero":
            raise ValueError(
                "`sr` can only be None if signal(s) are "
                "length 1 AND `ic` is not 'zero'"
            )
        sr = 1.0  # can be anything, just needed for calculations

    parallel, ncpu = _process_parallel(parallel, LF, N * H, maxcpu, getresp)

    if parallel == "yes":
        # global shared vars will be:
        # SRSmax_, WN_, HIST_, SIG_, ICVALS_
        SRSmax = (createSharedArray((LF, H)), (LF, H))
        WN = (copyToSharedArray(wn), wn.shape)
        HIST = (None, None)
    else:
        SRSmax = np.empty((LF, H))

    sig, s1, doic, icvals = _process_ic(sig, ic, stype)

    mf = np.max(freq)
    if rolloff == "prefilter":
        sig, sr = rollfunc(sig, sr, ppc, mf)
        rollfunc = None  # rolloff = 'none'

    if rollfunc and mf != 0 and sr / mf < ppc:
        sig, sr = rollfunc(sig, sr, ppc, mf)
        N = sig.shape[0]
    rollfunc = None

    M = N
    if ptr:
        sig, N = _add_one_cycle(sig, freq, sr, H, ic, s1)

    if getresp:
        resp = {}
        resp["sr"] = sr
        # hist is:  len(time) x nsignals x len(freq)
        if ptr == 2:
            # residual
            resp["t"] = np.arange(M, N) / sr
            if parallel == "yes":
                HIST = (createSharedArray((N - M, H, LF)), (N - M, H, LF))
            else:
                resp["hist"] = np.empty((N - M, H, LF))
        else:
            resp["t"] = np.arange(N) / sr
            if parallel == "yes":
                HIST = (createSharedArray((N, H, LF)), (N, H, LF))
            else:
                resp["hist"] = np.empty((N, H, LF))

    # S is starting time for calcs; only non-zero if residual only:
    S = M if ptr == 2 else 0

    if doic:
        if parallel == "yes":
            SIG = (copyToSharedArray(sig), sig.shape)
            ICVALS = (copyToSharedArray(icvals), icvals.shape)
            args = (coeffunc, Q, 1 / sr, methfunc, S, stype)
            gvars = (WN, SIG, ICVALS, SRSmax, HIST)
            func = _dosrs_ic if getresp else _dosrs_nohist_ic
            with mp.Pool(
                processes=ncpu, initializer=_mk_par_globals_ic, initargs=gvars
            ) as pool:
                for _ in pool.imap_unordered(func, zip(range(LF), it.repeat(args, LF))):
                    pass
            SRSmax = np.frombuffer(SRSmax[0]).reshape(SRSmax[1])
            if getresp:
                HIST = np.frombuffer(HIST[0]).reshape(HIST[1])
                resp["hist"] = HIST
        else:
            dT = 1 / sr
            for j in range(LF):
                b, a = coeffunc(Q, dT, wn[j])
                resphist = signal.lfilter(b, a, sig, axis=0)
                if stype == "reldisp":
                    resphist += icvals / wn[j] ** 2
                elif stype == "pvelo":
                    resphist += icvals / wn[j]
                else:
                    # stype == 'pacce' or 'absacce'
                    resphist += icvals
                SRSmax[j] = methfunc(resphist[S:])
                if getresp:
                    resp["hist"][:, :, j] = resphist[S:]
    else:
        # no initial conditions to worry about:
        if parallel == "yes":
            SIG = (copyToSharedArray(sig), sig.shape)
            args = (coeffunc, Q, 1 / sr, methfunc, S)
            gvars = (WN, SIG, SRSmax, HIST)
            func = _dosrs if getresp else _dosrs_nohist
            with mp.Pool(
                processes=ncpu, initializer=_mk_par_globals, initargs=gvars
            ) as pool:
                for _ in pool.imap_unordered(func, zip(range(LF), it.repeat(args, LF))):
                    pass
            SRSmax = np.frombuffer(SRSmax[0]).reshape(SRSmax[1])
            if getresp:
                HIST = np.frombuffer(HIST[0]).reshape(HIST[1])
                resp["hist"] = HIST
        else:
            dT = 1 / sr
            for j in range(LF):
                b, a = coeffunc(Q, dT, wn[j])
                resphist = signal.lfilter(b, a, sig, axis=0)
                SRSmax[j] = methfunc(resphist[S:])
                if getresp:
                    resp["hist"][:, :, j] = resphist[S:]
    if oneD:
        SRSmax = SRSmax.ravel()
    if getresp:
        if eqsine:
            SRSmax /= Q
            resp["hist"] /= Q
        return SRSmax, resp
    if eqsine:
        SRSmax /= Q
    return SRSmax


def vrs(spec, freq, Q, linear, Fn=None, getmiles=False, getresp=False):
    r"""
    Vibration response specturm - RMS response of single DOF systems
    to base PSD(s).

    Parameters
    ----------
    spec : 2d ndarray or 2-element tuple/list
        If ndarray, its columns are ``[Freq, PSD1, PSD2, ... PSDn]``.
        Otherwise, it must be a 2-element tuple or list, eg:
        ``(Freq, PSD)`` where PSD is: ``[PSD1, PSD2, ... PSDn]``. In
        the second usage, PSD can be 1d in which case the outputs will
        also be 1d; in the first usage, the outputs will be 2d. The
        frequency vector must be monotonically increasing.
    freq : 1d array_like
        Vector of frequencies to define the integration step; see
        usage note 2 below.
    Q : scalar
        Dynamic amplification factor :math:`Q = 1/(2\zeta)` where
        :math:`\zeta` is the fraction of critical damping.
    linear : bool
        If True, use linear interpolation to expand `spec` to the
        frequencies in `freq`. If False, `spec` is expanded via
        interpolation in log space. In other words:

        ================   ==========================================
        Use:               When:
        ================   ==========================================
        ``linear=False``   `spec` is an actual PSD test specification
                           -- that is, it uses constant db/octave
                           slopes
        ``linear=True``    `spec` doesn't use constant db/octave
                           slopes (eg, an analysis curve)
        ================   ==========================================

        The :func:`pyyeti.psd.interp` routine is called to perform the
        interpolation.
    Fn : 1d array or None
        Defines the frequency(s) at which to compute the response. If
        None, ``Fn = freq``.
    getmiles : bool
        If True, return the Miles' equation estimate for the RMS
        response. Miles' equation assumes the PSD is flat and extends
        forever in both directions from frequency. The estimate is
        typically good if the PSD is flat for at least 2 octaves in
        both directions and the damping is less that 10%.
    getresp : bool
        If True, return the PSD response curves at the frequency(s) in
        `Fn`. Note: internally, this will also set `getmiles` to
        True.

    Returns
    -------
    Zvrs : 1d or 2d ndarray
        The SDOF RMS acceleration response spectrum. For example, the
        response spectrum is in Grms if the spec is given in G^2/Hz.
        ``Zvrs.shape = (len(Fn), n)``, where n is the number of input
        specifications. Will be 1d only if the PSD input in `spec` was
        1d (in the second usage); in that case,
        ``Zvrs.shape = (len(Fn),)``.
    Zmiles : 1d or 2d ndarray; optional
        The Miles' estimate of `Zvrs`. Only returned if `getmiles` or
        `getresp` is True. ``Zmiles(f) = sqrt(pi/2*Fn*Q*psd(f))``.
        Shape will be the same as `Zvrs`.
    PSDresp : dictionary; optional
        Only returned if `getresp` is True. Members:

        ======   =====================================================
         key     value
        ======   =====================================================
        'f'      frequency vector for responses; same as `freq`
        'psd'    3-D array; shape = ``(len(Fn), n, len(f))``, where
                 `n` is the number of input specifications
        ======   =====================================================

    Notes
    -----
    VRS [#srs6]_ computes the acceleration RMS (root-mean-square)
    response of a spectrum of single DOF systems that are excited by
    an input base acceleration PSD(s)::

                      _____    ^
                     |     |   |
                     |  M  |  ---  SDOF acceleration PSD response
                     |_____|
                      /  |
                    K \ |_| C  ^
                      /  |     |
                    [=======] ---  input base acceleration PSD

    The response of each system is computed independently by
    integration across the entire frequency range as specified in
    `freq`.

    The equation for the VRS is:

    .. math::
        Z_{vrs}(f_n) = \sqrt { \sum\limits_{i} {
           \frac{1 + (p_i / Q)^2}
                {(1 - p_i^2)^2 + (p_i / Q)^2}}
           \cdot PSD(freq_i) \cdot \Delta freq_i}\;\;;\;
           p_i = \frac{freq_i}{f_n}

    Miles' equation is:

    .. math::
        Z_{miles}(f_n) = \sqrt{\frac{\pi}{2} \cdot f_n \cdot Q
        \cdot PSD(f_n)}

    Important usage notes:

    1. Responses calculated at the extreme frequency points may not be
       accurate because the transfer function is cut off during
       integration. As a rule of thumb, accuracy can be expected an
       octave away from the end points.

    2. The integration is not accurate until ``delta_f`` as computed
       from `freq` is less than ``f/Q``, i.e. the response at
       frequency ``f`` is not accurate unless ``delta_f < f/Q``, where
       ``Q=1/2/zeta``. The integration should be conservative if this
       condition is not met and ``delta_f`` is not unreasonably large.

    3. Applying a flat PSD spectrum can be used to determine if the
       ``delta_f`` is good since you can compare to the Miles'
       equation results.

    To estimate a peak from the RMS, consider using
    ``sqrt(2*log(f*T0))`` for the peak factor instead of the common 3
    (for 3-rms or 3-sigma). This peak factor is explained under the
    `resp_time` option in
    :func:`pyyeti.cla.DR_Results.psd_data_recovery`.

    See also
    --------
    :func:`srs`, :func:`srs_frf`, :func:`pyyeti.psd.interp`

    References
    ----------
    .. [#srs6] Tom Irvine, "An Introduction to the Vibration Response
           Spectrum - Revision D",
           http://www.vibrationdata.com/tutorials2/vrs.pdf

    Examples
    --------
    Compute response spectra for example shown in reference. The
    results should be::

        vrs                = [6.38, 11.09, 16.06]
        miles              = [6.47, 11.21, 15.04]
        response PSD peaks = [2.69,  4.04,  1.47]

    >>> import numpy as np
    >>> from pyyeti import srs
    >>> spec = np.array([[20, .0053],
    ...                  [150, .04],
    ...                  [600, .04],
    ...                  [2000, .0036]])
    >>> frq = np.arange(20, 2000, 2.)
    >>> Q = 10
    >>> fn = [100, 200, 1000]
    >>> v, m, resp = srs.vrs((spec[:, 0], spec[:, 1]), frq, Q,
    ...                      linear=False, Fn=fn, getresp=True)
    >>> np.set_printoptions(precision=2)
    >>> v
    array([  6.38,  11.09,  16.06])
    >>> m
    array([  6.47,  11.21,  15.04])
    >>> resp['psd'][:, 0].max(axis=1)
    array([ 2.69,  4.04,  1.47])
    """
    if Q <= 0.5:
        raise ValueError("Q must be > 0.5 since VRS assumes underdamped equations.")

    Freq, PSD, npsds = psd.proc_psd_spec(spec)
    freq = np.atleast_1d(freq)

    if Fn is None:
        Fn = freq
        do_interp = False
    else:
        Fn = np.atleast_1d(Fn)
        freq = np.unique(np.hstack((freq, Fn)))
        do_interp = True
    rf = len(freq)

    # expand PSD:
    psdfull = psd.interp((Freq, PSD), freq, linear)
    if PSD.ndim == 1:
        psdfull = psdfull[:, None]

    # check for adequate delta_freq for integration:
    df = np.empty(rf)
    df[:-1] = freq[1:] - freq[:-1]
    df[-1] = freq[-1] - freq[-2]
    if do_interp:
        pv = np.where((freq >= Fn.min()) & (freq <= Fn.max()))[0]
        bad_df = df[pv] > freq[pv] / Q
    else:
        bad_df = df > freq / Q

    if bad_df.any():
        warn(
            "Integration frequency vector may produce inaccurate results;"
            " refine the step. See the documentation for more information.",
            RuntimeWarning,
        )

    # Create delta_f for area calculation:
    df = np.empty(rf)
    df[1:-1] = (freq[2:] - freq[:-2]) / 2
    df[0] = freq[1] - freq[0]
    df[-1] = freq[-1] - freq[-2]

    # Compute Miles' equation
    if getresp or getmiles:
        if do_interp:
            ifunc = interp.interp1d(
                freq,
                psdfull,
                axis=0,
                bounds_error=False,
                fill_value=0,
                assume_sorted=True,
            )
            psdf2 = ifunc(Fn)
            z_miles = np.sqrt((np.pi / 2 * Fn * Q) * psdf2.T).T
        else:
            z_miles = np.sqrt((np.pi / 2 * freq * Q) * psdfull.T).T
        if PSD.ndim == 1:
            z_miles = z_miles.ravel()

    # Compute VRS at each frequency
    z_vrs = np.empty((len(Fn), npsds))
    zeta = 1 / 2 / Q
    if getresp:
        psd_vrs = np.empty((len(Fn), npsds, len(freq)))
        for i, fn in enumerate(Fn):
            p = freq / fn
            p2z2 = (2 * zeta * p) ** 2
            t = ((1 + p2z2) / ((1 - p ** 2) ** 2 + p2z2)) * psdfull.T
            psd_vrs[i] = t  # npsds x len(freq)
            z_vrs[i] = np.sqrt(np.sum(df * t, axis=1))
        resp = {}
        resp["f"] = freq
        resp["psd"] = psd_vrs
        if PSD.ndim == 1:
            z_vrs = z_vrs.ravel()
        return z_vrs, z_miles, resp

    for i, fn in enumerate(Fn):
        p = freq / fn
        p2z2 = (2 * zeta * p) ** 2
        t = ((1 + p2z2) / ((1 - p ** 2) ** 2 + p2z2) * df) * psdfull.T
        z_vrs[i] = np.sqrt(np.sum(t, axis=1))
    if PSD.ndim == 1:
        z_vrs = z_vrs.ravel()
    if getmiles:
        return z_vrs, z_miles
    return z_vrs


def srs_frf(frf, frf_frq, srs_frq, Q, getresp=False):
    r"""
    Compute SRS from frequency response functions.

    Parameters
    ----------
    frf : 2d array_like
        Columns of FRF data defining base motion in frequency domain:
        [FRF1, FRF2, ... FRFn].  The FRFs can be complex; if so, this
        routine uses the absolute value of each column before
        interpolating to a frequency vector that includes system
        frequencies (note that using the absolute value gives
        equivalent results).  Number of rows must equal
        ``len(frf_frq)``.  If it is 1d, it is reshaped into a single
        column: [FRF1].
    frf_frq : 1d array_like
        Frequency vector in Hz for the FRF data.
    srs_frq : 1d array_like
        Frequency vector in Hz for the SRS.
    Q : scalar
        Dynamic amplification factor :math:`Q = 1/(2\zeta)` where
        :math:`\zeta` is the fraction of critical damping.
    getresp : bool; optional
        If True, return the complex response frfs (see `resp` output
        below).

    Returns
    -------
    sh : 2d ndarray
        The SRS results: [SRS1, SRS2, .... SRSn];
        ``sh.shape = (len(srs_frq), n)`` where n is the number of
        FRFs.
    resp : dictionary; optional
        Only returned if `getresp` is True. Members:

        ======   =====================================================
         key     value
        ======   =====================================================
        'freq'   frequency vector for responses; this is a superset of
                 `frf_frq` and `srs_frf` with near-duplicates removed
        'frfs'   3-D array; shape = ``(len(freq), n, len(srs_frq))``
        ======   =====================================================

    Notes
    -----
    The shock response spectrum is the response of single DOF
    system(s) that are excited by an input base acceleration FRF::

                      _____    ^
                     |     |   |
                     |  M  |  ---  SDOF response (x)
                     |_____|
                      /  |
                    K \ |_| C  ^
                      /  |     |
                    [=======] ---  input base acceleration (frf)

    The response of each system is computed independently by
    integration across the entire frequency range as specified in
    ``resp["freq"]`` above.

    Derivation of the equation of motion follows. First, let:

    .. math::
        \begin{aligned}
        \ddot{z} &= sig \\
        M &= 1 \\
        K &= \omega^2 \\
        C &= 2\zeta\omega \\
        \end{aligned}

    Note that :math:`\omega=2 \pi f` where :math:`f` is a frequency in
    Hz from the input `srs_frq`. The equation of motion is:

    .. math::
        \begin{aligned}
        \ddot{x} &= \sum Forces\; on\; M \\
        &= \omega^2(z-x)+2\zeta\omega(\dot{z}-\dot{x})
        \end{aligned}

    Define a relative coordinate :math:`u = x - z`. Then:

    .. math::
        \begin{aligned}
        \ddot{x}+2\zeta\omega\dot{u}+\omega^2 u &= 0 \\
        \ddot{u}+2\zeta\omega\dot{u}+\omega^2 u &= -\ddot{z}
        \end{aligned}

    Using the Fourier transform :math:`\mathcal{F}[x(t)] = X(\Omega)`:

    .. math::
        \begin{aligned}
        (-\Omega^2+2\zeta\omega\Omega j + \omega^2) U(\Omega) &=
        -Z_{acce}(\Omega) \\
        U(\Omega) &= -Z_{acce}(\Omega) /
        (-\Omega^2+2\zeta\omega\Omega j + \omega^2) \\
        U_{acce}(\Omega) &= \Omega^2 Z_{acce}(\Omega) /
        (-\Omega^2+2\zeta\omega\Omega j + \omega^2)
        \end{aligned}

    Then:

    .. math::
        X_{acce}(\Omega) = U_{acce}(\Omega) + Z_{acce}(\Omega)

    The return value `sh` contains the peak of the absolute value of
    :math:`X_{acce}(\Omega)`.

    Examples
    --------
    Make up a complex FRF and compute the shock response spectrum
    (srs). The peak srs value is at the peak FRF value and can be
    shown to be: ``srs_peak = frf_peak * sqrt(Q**2 + 1)``.

    >>> from pyyeti import srs
    >>> import numpy as np
    >>> frf = np.array([1, 1, 1, 1, 1.2, 1.4, 2.6, 1, .7, .8, 1])
    >>> frf = frf + 1j * np.random.randn(len(frf))*.05
    >>> frf_frq = .5 * np.arange(len(frf)) + 1.
    >>> srs_frq = np.arange(.1, 8, .1)
    >>> Q = 20
    >>> sh = srs.srs_frf(frf, frf_frq, srs_frq, Q)
    >>> pk_should_be = np.abs(frf).max() * np.sqrt(Q**2+1)
    >>> np.abs(sh.max() - pk_should_be) < 1e-13
    True
    """
    srs_frq = np.asarray(srs_frq)
    ws = 2.0 * np.pi * srs_frq
    n = len(ws)
    ms = np.ones(n, float)
    bs = 1 / Q * ws
    ks = ws ** 2

    frf_frq = np.asarray(frf_frq)
    frf = np.asarray(frf)
    if frf.ndim == 1:
        frf = frf.reshape(-1, 1)
    nfrf = frf.shape[1]
    frf = np.abs(frf)

    # include ks frequencies (srs_frq) in the forcing function
    ffreq = np.sort(np.hstack((frf_frq, srs_frq)))
    df = np.diff(ffreq)
    pv = np.ones(len(ffreq), bool)
    pv[1:] = df > 1.0e-5
    ffreq = ffreq[pv]
    nf = len(ffreq)

    if len(frf_frq) == 1:
        newfrf = np.zeros((nf, nfrf), float)
        i = np.searchsorted(ffreq, frf_frq)
        if i == nf:
            i -= 1
        newfrf[i] = frf
        frf = newfrf
    else:
        ifunc = interp.interp1d(
            frf_frq, frf, axis=0, bounds_error=False, fill_value=0, assume_sorted=True
        )
        frf = ifunc(ffreq)

    shk = np.empty((n, nfrf), float)
    pvrb = ks < 0.005  # ks/ms < .005 ... since ms == 1
    pvel = np.logical_not(pvrb)
    rb = np.any(pvrb)
    el = np.any(pvel)

    # setup frequency scale for solution:
    freqw = 2 * np.pi * ffreq
    if el:
        fw = freqw.reshape(1, -1)
        H = (
            ks[pvel].reshape(-1, 1)
            - ms[pvel].reshape(-1, 1) @ fw ** 2
            + 1j * (bs[pvel].reshape(-1, 1) @ fw)
        )

    a = np.empty((n, nf), complex)
    if getresp:
        frfs = np.empty((nf, nfrf, n), complex)

    for j in range(nfrf):
        # compute relative response, then absolute (see eqns in srs)
        a[:] = 0.0
        fs = frf[:, j]  # len(frf)
        if rb:
            a[pvrb] = -fs  # / ms ... since ms == 1
        if el:
            a[pvel] = (fs * freqw ** 2) / H
        a += fs
        if getresp:
            frfs[:, j, :] = a.T
        shk[:, j] = abs(a).max(axis=1)

    if getresp:
        resp = {"freq": ffreq, "frfs": frfs}
        return shk, resp

    return shk


def srsmap(timeslice, tsoverlap, sig, sr, freq, Q, wep=0, **srsargs):
    r"""
    Make a shock response spectral map ('waterfall') over time and
    frequency.

    Parameters
    ----------
    timeslice : scalar or string-integer
        If scalar, it is the length in seconds for each slice. If
        string, it contains the integer number of points for each
        slice. For example, if `sr` is 1000 samples/second,
        ``timeslice=0.75`` is equivalent to ``timeslice="750"``.
    tsoverlap : scalar in [0, 1) or string-integer
        If scalar, is the fraction of each time-slice to overlap. If
        string, it contains the integer number of points to
        overlap. For example, if `sr` is 1000 samples/second,
        ``tsoverlap=0.5`` and ``tsoverlap="500"`` each specify 50%
        overlap.
    sig : 1d array_like
        Base acceleration signal; vector.
    sr : scalar
        The sample rate (samples/sec)
    freq : array_like
        Frequency vector in Hz. This defines the single DOF systems
        to use.
    Q : scalar
        Dynamic amplification factor :math:`Q = 1/(2\zeta)` where
        :math:`\zeta` is the fraction of critical damping.
    wep : scalar
        Argument for the :func:`pyyeti.dsp.windowends`; specifies the
        window-ends portion. Each time slice is passed through
        :func:`pyyeti.dsp.windowends` if wep > 0.
    **srsargs : miscellaneous options for :func:`srs`
        Allows the setting of `ic`, `stype`, `peak`, `eqsine`, etc
        options for :func:`srs`.  See :func:`srs` for more
        information.

    Returns
    -------
    mp : 2d ndarray
        The SRS map; columns span time, rows span frequency (so each
        column is an SRS curve). Time increases going across the
        columns and frequency increases going down the rows.
    t : 1d ndarray
        Time vector of center times; corresponds to columns in map.
        Signal is assumed to start at time = 0.
    f : 1d ndarray
        Frequency vector equal to the input `freq`; corresponds to
        rows in map.

    Notes
    -----
    This routine calls :func:`pyyeti.dsp.waterfall` for handling the
    timeslices and preparing the output.  :func:`srs` and
    :func:`pyyeti.dsp.windowends` are passed to that function.

    See also
    --------
    :func:`srs`, :func:`pyyeti.dsp.waterfall`,
    :func:`pyyeti.dsp.windowends`

    Examples
    --------
    Generate a sine sweep signal @ 4 octaves/min; process in 2-second
    windows with 50% overlap; 2% windowends, compute equivalent sine.

    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from pyyeti import srs
        >>> from pyyeti import ytools
        >>> from matplotlib import cm
        >>> sig, t, f = ytools.gensweep(10, 1, 50, 4)
        >>> sr = 1/t[1]
        >>> frq = np.arange(1., 50.1)
        >>> Q = 20
        >>> mp, t, f = srs.srsmap(2, .5, sig, sr, frq, Q, .02,
        ...                       eqsine=1)
        >>> _ = plt.figure('Example')
        >>> plt.clf()
        >>> _ = plt.contour(t, f, mp, 40, cmap=cm.plasma)
        >>> cbar = plt.colorbar()
        >>> cbar.filled = True
        >>> cbar.draw_all()
        >>> _ = plt.xlabel('Time (s)')
        >>> _ = plt.ylabel('Frequency (Hz)')
        >>> ttl = 'EQSINE Map of Sine-Sweep @ 4 oct/min, Q = 20'
        >>> _ = plt.title(ttl)

    Also show results on a 3D surface plot:

    .. plot::
        :context: close-figs

        >>> fig = plt.figure('Example 2')
        >>> fig.clf()
        >>> from mpl_toolkits.mplot3d import Axes3D
        >>> ax = fig.gca(projection='3d')
        >>> x, y = np.meshgrid(t, f)
        >>> surf = ax.plot_surface(x, y, mp, rstride=1, cstride=1,
        ...                        linewidth=0, cmap=cm.plasma)
        >>> _ = fig.colorbar(surf, shrink=0.5, aspect=5)
        >>> ax.view_init(azim=-123, elev=48)
        >>> _ = ax.set_xlabel('Time (s)')
        >>> _ = ax.set_ylabel('Frequency (Hz)')
        >>> _ = ax.set_zlabel('Amplitude')
        >>> _ = plt.title(ttl)
    """
    return dsp.waterfall(
        sig,
        sr,
        timeslice,
        tsoverlap,
        srs,
        which=None,
        freq=freq,
        kwargs=dict(sr=sr, freq=freq, Q=Q, **srsargs),
        slicefunc=dsp.windowends,
        slicekwargs=dict(portion=wep),
    )



# -*- coding: utf-8 -*-
"""
Tools for computing the matrix exponential and the integral.
"""

import warnings
import numpy as np
import scipy.sparse.linalg.matfuncs as mf
import scipy.linalg as la


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


class _ExpmIntPadeHelper(mf._ExpmPadeHelper):
    def __init__(self, A, structure=None, use_exact_onenorm=False):
        """
        Initialize the object.

        Parameters
        ----------
        A : a dense or sparse square numpy matrix or ndarray
            The matrix to be exponentiated.
        structure : str, optional
            A string describing the structure of matrix `A`.
            Only `upper_triangular` is currently supported.
        use_exact_onenorm : bool, optional
            If True then only the exact one-norm of matrix powers and
            products will be used. Otherwise, the one-norm of powers
            and products may initially be estimated.
        """
        mf._ExpmPadeHelper.__init__(self, A, structure=None, use_exact_onenorm=False)
        self._A3 = None
        self._A5 = None

    @property
    def A3(self):
        if self._A3 is None:
            self._A3 = mf._smart_matrix_product(
                self.A, self.A2, structure=self.structure
            )
        return self._A3

    @property
    def A5(self):
        if self._A5 is None:
            self._A5 = mf._smart_matrix_product(
                self.A, self.A4, structure=self.structure
            )
        return self._A5

    def pade3_i(self, h):
        b = (120.0, 60.0, 12.0, 1.0)
        U = b[3] * self.A3 + b[1] * self.A
        V = b[2] * self.A2 + b[0] * self.ident
        p = (840.0, 60.0, 20.0, 1.0)
        q = (840.0, -360.0, 60.0, -4.0)
        P = h * (p[3] * self.A3 + p[2] * self.A2 + p[1] * self.A + p[0] * self.ident)
        Q = q[3] * self.A3 + q[2] * self.A2 + q[1] * self.A + q[0] * self.ident
        return U, V, P, Q

    def pade5_i(self, h):
        b = (30240.0, 15120.0, 3360.0, 420.0, 30.0, 1.0)
        U = b[5] * self.A5 + b[3] * self.A3 + b[1] * self.A
        V = b[4] * self.A4 + b[2] * self.A2 + b[0] * self.ident
        p = (332640.0, 15120.0, 10080.0, 420.0, 42.0, 1.0)
        q = (332640.0, -151200.0, 30240.0, -3360.0, 210.0, -6.0)
        P = h * (
            p[5] * self.A5
            + p[4] * self.A4
            + p[3] * self.A3
            + p[2] * self.A2
            + p[1] * self.A
            + p[0] * self.ident
        )
        Q = (
            q[5] * self.A5
            + q[4] * self.A4
            + q[3] * self.A3
            + q[2] * self.A2
            + q[1] * self.A
            + q[0] * self.ident
        )
        return U, V, P, Q

    def pade7_i(self, h):
        U, V = self.pade7()
        p = (259459200.0, 8648640.0, 8648640.0, 277200.0, 55440.0, 1512.0, 72.0, 1.0)
        q = (
            259459200.0,
            -121080960.0,
            25945920.0,
            -3326400.0,
            277200.0,
            -15120.0,
            504.0,
            -8.0,
        )
        P = h * (
            mf._smart_matrix_product(
                self.A,
                p[7] * self.A6 + p[5] * self.A4 + p[3] * self.A2 + p[1] * self.ident,
                structure=self.structure,
            )
            + p[6] * self.A6
            + p[4] * self.A4
            + p[2] * self.A2
            + p[0] * self.ident
        )
        Q = (
            mf._smart_matrix_product(
                self.A,
                q[7] * self.A6 + q[5] * self.A4 + q[3] * self.A2 + q[1] * self.ident,
                structure=self.structure,
            )
            + q[6] * self.A6
            + q[4] * self.A4
            + q[2] * self.A2
            + q[0] * self.ident
        )
        return U, V, P, Q

    def pade9_i(self, h):
        U, V = self.pade9()
        p = (
            335221286400.0,
            8821612800.0,
            11762150400.0,
            302702400.0,
            90810720.0,
            2162160.0,
            205920.0,
            3960.0,
            110.0,
            1.0,
        )
        q = (
            335221286400.0,
            -158789030400.0,
            35286451200.0,
            -4843238400.0,
            454053600.0,
            -30270240.0,
            1441440.0,
            -47520.0,
            990.0,
            -10.0,
        )
        P = h * (
            mf._smart_matrix_product(
                self.A,
                p[9] * self.A8
                + p[7] * self.A6
                + p[5] * self.A4
                + p[3] * self.A2
                + p[1] * self.ident,
                structure=self.structure,
            )
            + p[8] * self.A8
            + p[6] * self.A6
            + p[4] * self.A4
            + p[2] * self.A2
            + p[0] * self.ident
        )
        Q = (
            mf._smart_matrix_product(
                self.A,
                q[9] * self.A8
                + q[7] * self.A6
                + q[5] * self.A4
                + q[3] * self.A2
                + q[1] * self.ident,
                structure=self.structure,
            )
            + q[8] * self.A8
            + q[6] * self.A6
            + q[4] * self.A4
            + q[2] * self.A2
            + q[0] * self.ident
        )
        return U, V, P, Q

    def pade13_scaled_i(self, s, h):
        b = (
            64764752532480000.0,
            32382376266240000.0,
            7771770303897600.0,
            1187353796428800.0,
            129060195264000.0,
            10559470521600.0,
            670442572800.0,
            33522128640.0,
            1323241920.0,
            40840800.0,
            960960.0,
            16380.0,
            182.0,
            1.0,
        )
        B = self.A * 2 ** -s
        h = h * 2 ** -s
        B2 = self.A2 * 2 ** (-2 * s)
        B4 = self.A4 * 2 ** (-4 * s)
        B6 = self.A6 * 2 ** (-6 * s)
        U2 = mf._smart_matrix_product(
            B6, b[13] * B6 + b[11] * B4 + b[9] * B2, structure=self.structure
        )
        U = mf._smart_matrix_product(
            B,
            (U2 + b[7] * B6 + b[5] * B4 + b[3] * B2 + b[1] * self.ident),
            structure=self.structure,
        )
        V2 = mf._smart_matrix_product(
            B6, b[12] * B6 + b[10] * B4 + b[8] * B2, structure=self.structure
        )
        V = V2 + b[6] * B6 + b[4] * B4 + b[2] * B2 + b[0] * self.ident
        p = (
            1748648318376960000.0,
            32382376266240000.0,
            64764752532480000.0,
            1187353796428800.0,
            593676898214400.0,
            10559470521600.0,
            2011327718400.0,
            33522128640.0,
            2793510720.0,
            40840800.0,
            1485120.0,
            16380.0,
            210.0,
            1.0,
        )
        q = (
            1748648318376960000.0,
            -841941782922240000.0,
            194294257597440000.0,
            -28496491114291200.0,
            2968384491072000.0,
            -232308351475200.0,
            14079294028800.0,
            -670442572800.0,
            25141596480.0,
            -735134400.0,
            16336320.0,
            -262080.0,
            2730.0,
            -14.0,
        )
        _P2 = mf._smart_matrix_product(
            B6, p[13] * B6 + p[11] * B4 + p[9] * B2, structure=self.structure
        )
        P2 = mf._smart_matrix_product(
            B,
            _P2 + p[7] * B6 + p[5] * B4 + p[3] * B2 + p[1] * self.ident,
            structure=self.structure,
        )
        P1 = mf._smart_matrix_product(
            B6, p[12] * B6 + p[10] * B4 + p[8] * B2, structure=self.structure
        )
        P = h * (P2 + P1 + p[6] * B6 + p[4] * B4 + p[2] * B2 + p[0] * self.ident)
        _Q2 = mf._smart_matrix_product(
            B6, q[13] * B6 + q[11] * B4 + q[9] * B2, structure=self.structure
        )
        Q2 = mf._smart_matrix_product(
            B,
            _Q2 + q[7] * B6 + q[5] * B4 + q[3] * B2 + q[1] * self.ident,
            structure=self.structure,
        )
        Q1 = mf._smart_matrix_product(
            B6, q[12] * B6 + q[10] * B4 + q[8] * B2, structure=self.structure
        )
        Q = Q2 + Q1 + q[6] * B6 + q[4] * B4 + q[2] * B2 + q[0] * self.ident
        return U, V, P, Q


def expmint(A, h, geti2=False):
    """
    Compute the matrix exponential and its integral(s) using Pade
    approximation.

    Parameters
    ----------
    A : (M, M) array_like or sparse matrix
        2D Array or Matrix (sparse or dense) to be exponentiated
    h : scalar
        Time step
    geti2 : bool
        If True, also return the `I2` integral below. Useful for
        first order holds.

    Returns
    -------
    E : (M, M) ndarray
        Matrix exponential of `A*h`: exp(A*h)
    I : (M, M) ndarray
        Integral of exp(A*t) dt from 0 to h
    I2 : (M, M) ndarray; optional
        Integral of exp(A*t)*t dt from 0 to h. Only returned if
        `geti2` is True.

    Notes
    -----
    This routine is modeled after and augments
    :func:`scipy.linalg.expm`. The power series expansions for these
    matrices are (I = identity):

    .. code-block:: none

      E = I + A*h + (A*h)**2/2! + (A*h)**3/3! + ...
      I1 = h*(I + A*h/2 + (A*h)**2/3! + (A*h)**3/4! + ...)
      I2 = h*h*(I/2 + A*h/3 + (A*h)**2/(4*2!) + (A*h)**3/(5*3!) + ...)

    If `A` is non-singular, the exact solutions for `I1` and `I2` are::

        E = exp(A*h)
        I1 = inv(A)*(E-I)
        I2 = inv(A)*(E*h-I1)

    The Pade approximants for those power series are used for `E` and
    `I1`. If necessary, the 'squaring and scaling' method will be used
    such that a 13th order Pade approximation will be accurate. See
    references [#exp1]_, [#exp2]_, and [#exp3]_ for more information.

    For `I2`, a Pade approximation is used if a 3rd, 5th, 7th or 9th
    order is accurate. If not, `A` is checked for singularity. If
    non-singular, the exact solution show above is used. If A is
    singular, a the power series is used directly until it converges
    (and a warning message is printed about using a finer time step.)

    References
    ----------
    .. [#exp1] Awad H. Al-Mohy and Nicholas J. Higham (2009)
           "A New Scaling and Squaring Algorithm for the Matrix
           Exponential."
           SIAM Journal on Matrix Analysis and Applications.
           31 (3). pp. 970-989. ISSN 1095-7162

    .. [#exp2] Nicholas J. Higham (2005)
           "The Scaling and Squaring Method for the Matrix Exponential
           Revisited."
           SIAM Journal on Matrix Analysis and Applications.
           Vol 26, No. 4, pp 1179-1193.

    .. [#exp3] David Westreich (1990)
           "A Practical Method for Computing the Exponential of a Matrix
           and its Integral."
           Communications in Applied Numerical Methods, Vol 6, 375-380.

    Examples
    --------
    >>> from pyyeti import expmint
    >>> import numpy as np
    >>> import scipy.linalg as la
    >>> np.set_printoptions(4)
    >>> a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> e, i, i2 = expmint.expmint(a, .05, True)
    >>> e
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> i
    array([[ 0.052 ,  0.0034,  0.0048],
           [ 0.0067,  0.0583,  0.01  ],
           [ 0.0114,  0.0133,  0.0651]])
    >>> i2
    array([[ 0.0013,  0.0001,  0.0002],
           [ 0.0002,  0.0015,  0.0003],
           [ 0.0004,  0.0005,  0.0018]])
    """
    # Avoid indiscriminate asarray() to allow sparse or other strange
    # arrays.
    if isinstance(A, (list, tuple)):
        A = np.asarray(A)
    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError("expected a square matrix")

    # Detect upper triangularity.
    if mf._is_upper_triangular(A):
        structure = mf.UPPER_TRIANGULAR
    else:
        structure = None

    # Hardcode a matrix order threshold for exact vs. estimated
    # one-norms.
    use_exact_onenorm = A.shape[0] < 200

    # Track functions of A to help compute the matrix exponential.
    H = _ExpmIntPadeHelper(
        A * h, structure=structure, use_exact_onenorm=use_exact_onenorm
    )

    def Return(U, V, P, Q, geti2, pade):
        E = mf._solve_P_Q(U, V, structure=structure)
        I = _solve_P_Q_2(P, Q, structure=structure)
        if geti2:
            return E, I, _geti2(H, E, I, h, pade)
        return E, I

    # Try Pade order 3.
    eta_1 = max(H.d4_loose, H.d6_loose)
    if eta_1 < 1.495585217958292e-002 and mf._ell(H.A, 3) == 0:
        U, V, P, Q = H.pade3_i(h)
        return Return(U, V, P, Q, geti2, 3)

    # Try Pade order 5.
    eta_2 = max(H.d4_tight, H.d6_loose)
    if eta_2 < 2.539398330063230e-001 and mf._ell(H.A, 5) == 0:
        U, V, P, Q = H.pade5_i(h)
        return Return(U, V, P, Q, geti2, 5)

    # Try Pade orders 7 and 9.
    eta_3 = max(H.d6_tight, H.d8_loose)
    if eta_3 < 9.504178996162932e-001 and mf._ell(H.A, 7) == 0:
        U, V, P, Q = H.pade7_i(h)
        return Return(U, V, P, Q, geti2, 7)

    if eta_3 < 2.097847961257068e000 and mf._ell(H.A, 9) == 0:
        U, V, P, Q = H.pade9_i(h)
        return Return(U, V, P, Q, geti2, 9)

    # Use Pade order 13.
    eta_4 = max(H.d8_loose, H.d10_loose)
    eta_5 = min(eta_3, eta_4)
    theta_13 = 4.25
    s = max(int(np.ceil(np.log2(eta_5 / theta_13))), 0)
    s = s + mf._ell(2 ** -s * H.A, 13)
    U, V, P, Q = H.pade13_scaled_i(s, h)
    E = mf._solve_P_Q(U, V, structure=structure)
    I = _solve_P_Q_2(P, Q, structure=structure)
    # E = r_13(A)^(2^s) by repeated squaring.
    for _ in range(s):
        I += I.dot(E)
        E = E.dot(E)
    if geti2:
        return E, I, _geti2(H, E, I, h, 13)
    return E, I


def _geti2(H, E, I, h, pade):
    """
    A helper routine for expmint; computes I2.  See :func:`expmint`.
    """
    if pade <= 3:
        p = (450240.0, 104860.0, 15480.0, 1001.0)
        q = (900480.0, -390600.0, 66240.0, -4540.0)
        P = (h * h) * (p[3] * H.A3 + p[2] * H.A2 + p[1] * H.A + p[0] * H.ident)
        Q = q[3] * H.A3 + q[2] * H.A2 + q[1] * H.A + q[0] * H.ident
        return _solve_P_Q_2(P, Q, H.structure)

    if pade <= 5:
        p = (
            69363423360.0,
            14569133040.0,
            2595781440.0,
            239983884.0,
            14154000.0,
            398959.0,
        )
        q = (
            138726846720.0,
            -63346298400.0,
            12740716800.0,
            -1425725280.0,
            89937120.0,
            -2602278.0,
        )
        P = (h * h) * (
            mf._smart_matrix_product(
                H.A, p[5] * H.A4 + p[3] * H.A2 + p[1] * H.ident, structure=H.structure
            )
            + p[4] * H.A4
            + p[2] * H.A2
            + p[0] * H.ident
        )
        Q = (
            mf._smart_matrix_product(
                H.A, q[5] * H.A4 + q[3] * H.A2 + q[1] * H.ident, structure=H.structure
            )
            + q[4] * H.A4
            + q[2] * H.A2
            + q[0] * H.ident
        )
        return _solve_P_Q_2(P, Q, H.structure)

    if pade <= 7:
        p = (
            41840477770291200.0,
            8321436219096000.0,
            1617627908856960.0,
            159144893827920.0,
            12344957190720.0,
            614984330664.0,
            19815037200.0,
            312129649.0,
        )
        q = (
            83680955540582400.0,
            -39144431255529600.0,
            8411304436254720.0,
            -1081869058670400.0,
            90503101180800.0,
            -4959510549840.0,
            166265789760.0,
            -2658297528.0,
        )
        P = (h * h) * (
            mf._smart_matrix_product(
                H.A,
                p[7] * H.A6 + p[5] * H.A4 + p[3] * H.A2 + p[1] * H.ident,
                structure=H.structure,
            )
            + p[6] * H.A6
            + p[4] * H.A4
            + p[2] * H.A2
            + p[0] * H.ident
        )
        Q = (
            mf._smart_matrix_product(
                H.A,
                q[7] * H.A6 + q[5] * H.A4 + q[3] * H.A2 + q[1] * H.ident,
                structure=H.structure,
            )
            + q[6] * H.A6
            + q[4] * H.A4
            + q[2] * H.A2
            + q[0] * H.ident
        )
        return _solve_P_Q_2(P, Q, H.structure)

    if pade <= 9:
        p = (
            69491589579005577984000.0,
            13362425784392593708800.0,
            2733673676598876211200.0,
            274218651095712523200.0,
            24016200124299102720.0,
            1393079626219366800.0,
            63177322465033920.0,
            1972643393629480.0,
            40190548856040.0,
            403978495031.0,
        )
        q = (
            138983179158011155968000.0,
            -65930601203222249894400.0,
            14675286699176463360000.0,
            -2017982140321398451200.0,
            189583633133409715200.0,
            -12669377873982429600.0,
            604988181888330240.0,
            -20009929231749600.0,
            418494695659920.0,
            -4247085597370.0,
        )
        P = (h * h) * (
            mf._smart_matrix_product(
                H.A,
                p[9] * H.A8 + p[7] * H.A6 + p[5] * H.A4 + p[3] * H.A2 + p[1] * H.ident,
                structure=H.structure,
            )
            + p[8] * H.A8
            + p[6] * H.A6
            + p[4] * H.A4
            + p[2] * H.A2
            + p[0] * H.ident
        )
        Q = (
            mf._smart_matrix_product(
                H.A,
                q[9] * H.A8 + q[7] * H.A6 + q[5] * H.A4 + q[3] * H.A2 + q[1] * H.ident,
                structure=H.structure,
            )
            + q[8] * H.A8
            + q[6] * H.A6
            + q[4] * H.A4
            + q[2] * H.A2
            + q[0] * H.ident
        )
        return _solve_P_Q_2(P, Q, H.structure)

    # second, try direct solution:
    n = H.A.shape[0]
    with warnings.catch_warnings():
        warnings.simplefilter("error", RuntimeWarning)
        try:
            lup = la.lu_factor(H.A)
            # decomp maybe worked ... check it:
            # I = A\(expm(A*h) - Ident)
            I_test = la.lu_solve(lup, h * (E - np.eye(n)))
            if np.allclose(I_test, I):
                # I2 = A\(expm(A*h)*h - A\(expm(A*h) - Ident))
                return la.lu_solve(lup, h * (E * h - I_test))
        except RuntimeWarning:
            pass

    # finally, use taylor series:
    msg = (
        "Using power series expansion directly for `I2` (see"
        " `expmint`).\nRecommendation: use a finer time step."
    )
    warnings.warn(msg, RuntimeWarning)
    I2 = np.eye(n) / 2
    term = H.A
    j = 1.0
    tol = 1e-15
    maxloops = 200
    while abs(term).max() > tol * abs(E).max() and j < maxloops:
        j += 1.0
        I2 += term / (j + 1)
        term = term.dot(H.A) / j
    if j >= maxloops:
        raise RuntimeError(
            f"maximum loops ({maxloops}) exceeded for power series expansion"
        )
    return (h * h) * I2


def _solve_P_Q_2(P, Q, structure=None):
    """
    A helper function for expmint -- modified from the scipy source.

    Parameters
    ----------
    P : ndarray
        Pade numerator.
    Q : ndarray
        Pade denominator.
    structure : str, optional
        A string describing the structure of both matrices `P` and
        `Q`.  Only `upper_triangular` is currently supported.

    Notes
    -----
    The `structure` argument is inspired by similar args
    for theano and cvxopt functions.

    """
    if mf.isspmatrix(P):
        return mf.spsolve(Q, P)
    elif structure is None:
        return mf.solve(Q, P)
    elif structure == mf.UPPER_TRIANGULAR:
        return mf.solve_triangular(Q, P)
    else:
        raise ValueError("unsupported matrix structure: " + str(structure))


def expmint_pow(A, h):
    """
    Compute the matrix exponential and its integrals using the power
    series expansion.

    Parameters
    ----------
    A : (M, M) array_like or sparse matrix
        2D Array or Matrix (sparse or dense) to be exponentiated
    h : scalar
        Time step

    Returns
    -------
    E : (M, M) ndarray
        Matrix exponential of `A*h`: exp(A*h)
    I : (M, M) ndarray
        Integral of exp(A*t) dt from 0 to h
    I2 : (M, M) ndarray
        Integral of exp(A*t)*t dt from 0 to h

    Notes
    -----
    This routine is a simple brute-force alternative to the more
    elegant and optimized :func:`expmint`. The power series expansions
    for these matrices are (I = identity):

    .. code-block:: none

      E = I + A*h + (A*h)**2/2! + (A*h)**3/3! + ...
      I1 = h*(I + A*h/2 + (A*h)**2/3! + (A*h)**3/4! + ...)
      I2 = h*h*(I/2 + A*h/3 + (A*h)**2/(4*2!) + (A*h)**3/(5*3!) + ...)

    Examples
    --------
    >>> from pyyeti import expmint
    >>> import numpy as np
    >>> import scipy.linalg as la
    >>> np.set_printoptions(4)
    >>> a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> e, i, i2 = expmint.expmint_pow(a, .05)
    >>> e
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])

    >>> i
    array([[ 0.052 ,  0.0034,  0.0048],
           [ 0.0067,  0.0583,  0.01  ],
           [ 0.0114,  0.0133,  0.0651]])

    >>> i2
    array([[ 0.0013,  0.0001,  0.0002],
           [ 0.0002,  0.0015,  0.0003],
           [ 0.0004,  0.0005,  0.0018]])

    """
    Ah = A * h
    E = np.eye(A.shape[0])
    Int1 = E.copy()
    Int2 = E / 2
    term = Ah
    j = 1.0
    tol = 1e-15
    maxloops = 200
    while abs(term).max() > tol * abs(E).max() and j < maxloops:
        j += 1.0
        E += term
        Int1 += term / j
        Int2 += term / (j + 1.0)
        term = term.dot(Ah) / j
    if j >= maxloops:
        raise RuntimeError(
            f"maximum loops ({maxloops}) exceeded for power series expansion"
        )
    return E, h * Int1, h * h * Int2


def _procBhalf(E, P, Q, order, B, half):
    """Helper function for getEPQ1 and getEPQ_pow"""
    if B is not None:
        P = P.dot(B)
        if order == 1:
            Q = Q.dot(B)
    elif half:
        n = P.shape[1]
        if n & 1:
            raise ValueError(
                "`A` must have an even number of rows/cols (or use ``half=False``"
            )
        n = n // 2
        P = P[:, :n]
        if order == 1:
            Q = Q[:, :n]
    return E, P, Q


def getEPQ1(A, h, order=1, B=None, half=False):
    """
    Returns E, P, Q for the exponential solver given the state-space
    matrix `A`.

    Parameters
    ----------
    A : 2d ndarray
        The state-space matrix:  ``xdot = A x + B u``
    h : scalar
        Time step.
    order : integer, optional

        - 0 for the zero order hold (force stays constant across
          time step)
        - 1 for the 1st order hold (force can vary linearly across
          time step)

    B : d2 ndarray or None; optional
        If array, it multiplies the inputs; if None, it is assumed
        identity.
    half : bool; optional
        If `B` is a 2d ndarray, `half` is ignored. Otherwise, if
        `half` is False, a full size identity (same size as `A`) is
        used for `B`. If `half` is True, only the first half of the
        columns are retained (which is handy for converting a 2nd
        order ODE into a 1st order ODE as
        :class:`pyyeti.ode.SolveExp2` does -- where there are force
        inputs only for the first half of the equations).

    Returns
    -------
    E, P, Q : 2d ndarrays, except if ``order == 0``, ``Q = 0.``
        These are the coefficient matrices used to solve the ODE::

            for j in range(nt):
                d[:, j+1] = E*d[:, j] + P*F[:, j] + Q*F[:, j+1]

    Notes
    -----
    Normally, :func:`getEPQ` would be called and that routine will
    call this one or :func:`getEPQ2`.

    `E` is the matrix exponential ``exp(A*h)`` and `P` and `Q` are
    functions of the integral(s) of the matrix exponential::

            if order == 1:
                E, I, I2 = expmint(A, h, 1)
                P = (I2/h) @ B
                Q = (I - I2/h) @ B
            else:
                E, I = expmint(A, h)
                P = I @ B
                Q = 0.

    See also
    --------
    :func:`expmint`, :class:`pyyeti.ode.SolveExp1`,
    :class:`pyyeti.ode.SolveExp2`

    Examples
    --------
    >>> from pyyeti import expmint
    >>> import numpy as np
    >>> import scipy.linalg as la
    >>> np.set_printoptions(4)
    >>> A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> B = np.array([[0, 1, 0]]).T
    >>> E, P, Q = expmint.getEPQ1(A, .05, order=1, B=B)
    >>> E
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> P
    array([[ 0.0024],
           [ 0.0308],
           [ 0.0091]])
    >>> Q
    array([[ 0.0011],
           [ 0.0276],
           [ 0.0041]])
    >>> E, P, Q = expmint.getEPQ1(A, .05, order=0, B=B)
    >>> E
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> P
    array([[ 0.0034],
           [ 0.0583],
           [ 0.0133]])
    >>> Q
    0.0
    """
    if order == 1:
        E, I, I2 = expmint(A, h, 1)
        P = I2 / h
        Q = I - P
    else:
        E, P = expmint(A, h)
        Q = 0.0
    return _procBhalf(E, P, Q, order, B, half)


def getEPQ_pow(A, h, order=1, B=None, half=False):
    """
    Returns E, P, Q for the exponential solver given the state-space
    matrix `A`. Uses the power series expansion directly.

    Parameters
    ----------
    A : 2d ndarray
        The state-space matrix:  ydotD - A y = f
    h : scalar
        Time step.
    order : integer, optional

        - 0 for the zero order hold (force stays constant across time
          step)
        - 1 for the 1st order hold (force can vary linearly across
          time step)

    B : d2 ndarray or None; optional
        If array, it multiplies the inputs; if None, it is assumed
        identity.
    half : bool; optional
        If `B` is a 2d ndarray, `half` is ignored. Otherwise, if
        `half` is False, a full size identity (same size as `A`) is
        used for `B`. If `half` is True, only the first half of the
        columns are retained (which is handy for converting a 2nd
        order ODE into a 1st order ODE as
        :class:`pyyeti.ode.SolveExp2` does -- where there are force
        inputs only for the first half of the equations).

    Returns
    -------
    E, P, Q : 2d ndarrays, except if ``order == 0``, ``Q = 0.``
        These are the coefficient matrices used to solve the ODE::

            for j in range(nt):
                d[:, j+1] = E*d[:, j] + P*F[:, j] + Q*F[:, j+1]

    Notes
    -----
    This routine is primarily for testing purposes; in general, use
    :func:`getEPQ` instead.

    `E` is the matrix exponential ``exp(A*h)`` and `P` and `Q` are
    functions of the integral(s) of the matrix exponential::

            if order == 1:
                E, I, I2 = expmint(A, h, 1)
                P = (I2/h) @ B
                Q = (I - I2/h) @ B
            else:
                E, I = expmint(A, h)
                P = I @ B
                Q = 0.

    See also
    --------
    :func:`getEPQ`, :func:`expmint_pow`,
    :class:`pyyeti.ode.SolveExp1`, :class:`pyyeti.ode.SolveExp2`

    Examples
    --------
    >>> from pyyeti import expmint
    >>> import numpy as np
    >>> import scipy.linalg as la
    >>> np.set_printoptions(4)
    >>> A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> B = np.array([[0, 1, 0]]).T
    >>> E, P, Q = expmint.getEPQ_pow(A, .05, order=1, B=B)
    >>> E
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> P
    array([[ 0.0024],
           [ 0.0308],
           [ 0.0091]])
    >>> Q
    array([[ 0.0011],
           [ 0.0276],
           [ 0.0041]])
    >>> E, P, Q = expmint.getEPQ_pow(A, .05, order=0, B=B)
    >>> E
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> P
    array([[ 0.0034],
           [ 0.0583],
           [ 0.0133]])
    >>> Q
    0.0
    """
    E, I, I2 = expmint_pow(A, h)
    if order == 1:
        P = I2 / h
        Q = I - P
    else:
        P = I
        Q = 0.0
    return _procBhalf(E, P, Q, order, B, half)


class _ExpmPadeHelper_SS(mf._ExpmPadeHelper):
    """
    Class derived from scipy.sparse.linalg.matfuncs._ExpmPadeHelper
    to assist in calculating the matrix exponential and integrals for
    a state-space matrix. See :func:`getEPQ`.
    """

    def __init__(self, A, ssA, order):
        """
        Initialize the object.

        Parameters
        ----------
        A : square ndarray
            The matrix to be exponentiated; formed in such a way to
            get the matrix exponential of `ssA` and its integrals.
            See :func:`getEPQ`.
        ssA : square ndarray
            State-space 'A' matrix; `A` has `ssA` in upper left.
        order : integer
            Specifies how to handle forces: ``order=1`` for first
            order hold; ``order=0`` for zero order hold.
        """
        mf._ExpmPadeHelper.__init__(self, A, structure=None, use_exact_onenorm=True)
        self.ssA = ssA
        self.order = order
        self.nss = ssA.shape[0]
        self.shape = (self.nss, A.shape[1])

    @property
    def A2(self):
        if self._A2 is None:
            n = self.nss
            if self.order == 0:
                self._A2 = np.dot(self.ssA, self.A[:n])
            else:
                i = (self.A.shape[1] - n) // 2
                self._A2 = np.zeros(self.shape, float)
                self._A2[:, : n + i] = np.dot(self.ssA, self.A[:n, : n + i])
                self._A2[:, n + i :] = self.A[:n, n : n + i]
        return self._A2

    @property
    def A4(self):
        if self._A4 is None:
            n = self.nss
            self._A4 = np.dot(self.A2[:, :n], self.A2)
        return self._A4

    @property
    def A6(self):
        if self._A6 is None:
            n = self.nss
            self._A6 = np.dot(self.A4[:, :n], self.A2)
        return self._A6

    @property
    def A8(self):
        if self._A8 is None:
            n = self.nss
            self._A8 = np.dot(self.A6[:, :n], self.A2)
        return self._A8

    @property
    def A10(self):
        if self._A10 is None:
            n = self.nss
            self._A10 = np.dot(self.A4[:, :n], self.A6)
        return self._A10

    def pade3(self):
        b = (120.0, 60.0, 12.0, 1.0)
        n = self.nss
        U = b[1] * self.A
        V = b[0] * self.ident
        U[:n] += np.dot(self.ssA, b[3] * self.A2)
        V[:n] += b[2] * self.A2
        return U, V

    def pade5(self):
        b = (30240.0, 15120.0, 3360.0, 420.0, 30.0, 1.0)
        n = self.nss
        U = b[1] * self.A
        V = b[0] * self.ident
        U[:n] += np.dot(self.ssA, b[5] * self.A4 + b[3] * self.A2)
        V[:n] += b[4] * self.A4 + b[2] * self.A2
        return U, V

    def pade7(self):
        b = (17297280.0, 8648640.0, 1995840.0, 277200.0, 25200.0, 1512.0, 56.0, 1.0)
        n = self.nss
        U = b[1] * self.A
        V = b[0] * self.ident
        U[:n] += np.dot(self.ssA, (b[7] * self.A6 + b[5] * self.A4 + b[3] * self.A2))
        V[:n] += b[6] * self.A6 + b[4] * self.A4 + b[2] * self.A2
        return U, V

    def pade9(self):
        b = (
            17643225600.0,
            8821612800.0,
            2075673600.0,
            302702400.0,
            30270240.0,
            2162160.0,
            110880.0,
            3960.0,
            90.0,
            1.0,
        )
        n = self.nss
        U = b[1] * self.A
        V = b[0] * self.ident
        U[:n] += np.dot(
            self.ssA,
            (b[9] * self.A8 + b[7] * self.A6 + b[5] * self.A4 + b[3] * self.A2),
        )
        V[:n] += b[8] * self.A8 + b[6] * self.A6 + b[4] * self.A4 + b[2] * self.A2
        return U, V

    def pade13_scaled(self, s):
        b = (
            64764752532480000.0,
            32382376266240000.0,
            7771770303897600.0,
            1187353796428800.0,
            129060195264000.0,
            10559470521600.0,
            670442572800.0,
            33522128640.0,
            1323241920.0,
            40840800.0,
            960960.0,
            16380.0,
            182.0,
            1.0,
        )
        B = self.A * 2 ** -s
        B2 = self.A2 * 2 ** (-2 * s)
        B4 = self.A4 * 2 ** (-4 * s)
        B6 = self.A6 * 2 ** (-6 * s)
        n = self.nss
        U = b[1] * B
        V = b[0] * self.ident
        U2 = np.dot(B6[:, :n], b[13] * B6 + b[11] * B4 + b[9] * B2)
        V2 = np.dot(B6[:, :n], b[12] * B6 + b[10] * B4 + b[8] * B2)
        U[:n] += np.dot(B[:n, :n], U2 + b[7] * B6 + b[5] * B4 + b[3] * B2)
        V[:n] += V2 + b[6] * B6 + b[4] * B4 + b[2] * B2
        return U, V


def _expm_SS(A, ssA, order):  # , use_exact_onenorm='auto'):
    # Track functions of A to help compute the matrix exponential.
    h = _ExpmPadeHelper_SS(A, ssA, order)
    structure = None

    # Try Pade order 3.
    eta_1 = max(h.d4_loose, h.d6_loose)
    if eta_1 < 1.495585217958292e-002 and mf._ell(h.A, 3) == 0:
        U, V = h.pade3()
        return mf._solve_P_Q(U, V, structure=structure)

    # Try Pade order 5.
    eta_2 = max(h.d4_tight, h.d6_loose)
    if eta_2 < 2.539398330063230e-001 and mf._ell(h.A, 5) == 0:
        U, V = h.pade5()
        return mf._solve_P_Q(U, V, structure=structure)

    # Try Pade orders 7 and 9.
    eta_3 = max(h.d6_tight, h.d8_loose)
    if eta_3 < 9.504178996162932e-001 and mf._ell(h.A, 7) == 0:
        U, V = h.pade7()
        return mf._solve_P_Q(U, V, structure=structure)
    if eta_3 < 2.097847961257068e000 and mf._ell(h.A, 9) == 0:
        U, V = h.pade9()
        return mf._solve_P_Q(U, V, structure=structure)

    # Use Pade order 13.
    eta_4 = max(h.d8_loose, h.d10_loose)
    eta_5 = min(eta_3, eta_4)
    theta_13 = 4.25
    s = max(int(np.ceil(np.log2(eta_5 / theta_13))), 0)
    s = s + mf._ell(2 ** -s * h.A, 13)
    U, V = h.pade13_scaled(s)
    X = mf._solve_P_Q(U, V, structure=structure)
    # X = r_13(A)^(2^s) by repeated squaring.
    for _ in range(s):
        X = X.dot(X)
    return X


def getEPQ2(A, h, order=1, B=None, half=False):
    """
    Returns E, P, Q for the exponential solver given the state-space
    matrix `A`.

    Parameters
    ----------
    A : 2d ndarray
        The state-space matrix:  ``xdot = A x + B u``
    h : scalar
        Time step.
    order : integer, optional

        - 0 for the zero order hold (force stays constant across
          time step)
        - 1 for the 1st order hold (force can vary linearly across
          time step)

    B : d2 ndarray or None; optional
        If array, it multiplies the inputs; if None, it is assumed
        identity.
    half : bool; optional
        If `B` is a 2d ndarray, `half` is ignored. Otherwise, if
        `half` is False, a full size identity (same size as `A`) is
        used for `B`. If `half` is True, only the first half of the
        columns are retained (which is handy for converting a 2nd
        order ODE into a 1st order ODE as
        :class:`pyyeti.ode.SolveExp2` does -- where there are force
        inputs only for the first half of the equations).

    Returns
    -------
    E, P, Q : 2d ndarrays, except if ``order == 0``, ``Q = 0.``
        These are the coefficient matrices used to solve the ODE::

            for j in range(nt):
                d[:, j+1] = E*d[:, j] + P*F[:, j] + Q*F[:, j+1]

    Notes
    -----
    Normally, :func:`getEPQ` would be called and that routine will
    call this one or :func:`getEPQ1`.

    This routine is an alternative to :func:`getEPQ1` and is
    generally slower but more robust for large time steps. (If `B` has
    only a few columns, it could also be faster than
    :func:`getEPQ1`.) `E` is the matrix exponential ``exp(A*h)`` and
    `P` and `Q` are functions of the integral(s) of the matrix
    exponential. They are calculated as follows (text from
    :func:`scipy.signal.lsim`).

    If order == 0::

        Zero-order hold
        Algorithm: to integrate from time 0 to time dt, we solve
          xdot = A x + B u,  x(0) = x0
          udot = 0,          u(0) = u0.

        Solution is
          [ x(dt) ]       [ A*dt   B*dt ] [ x0 ]
          [ u(dt) ] = exp [  0     0    ] [ u0 ]

    The `E`, and `P` matrices are partitions of the matrix
    exponential and `Q` is zero.

    If order == 1::

        Linear interpolation between steps
        Algorithm: to integrate from time 0 to time dt, with
        linear interpolation between inputs u(0) = u0 and u(dt) = u1,
        we solve:
          xdot = A x + B u,        x(0) = x0
          udot = (u1 - u0) / dt,   u(0) = u0.

        Solution is
          [ x(dt) ]       [ A*dt  B*dt  0 ] [  x0   ]
          [ u(dt) ] = exp [  0     0    I ] [  u0   ]
          [u1 - u0]       [  0     0    0 ] [u1 - u0]

    The `E`, `P` and `Q` matrices are partitions of the matrix
    exponential.

    See also
    --------
    :func:`getEPQ1`, :func:`getEPQ_pow`,
    :class:`pyyeti.ode.SolveExp1`, :class:`pyyeti.ode.SolveExp2`

    Examples
    --------
    >>> from pyyeti import expmint
    >>> import numpy as np
    >>> import scipy.linalg as la
    >>> np.set_printoptions(4)
    >>> A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> B = np.array([[0, 1, 0]]).T
    >>> E, P, Q = expmint.getEPQ2(A, .05, order=1, B=B)
    >>> E
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> P
    array([[ 0.0024],
           [ 0.0308],
           [ 0.0091]])
    >>> Q
    array([[ 0.0011],
           [ 0.0276],
           [ 0.0041]])
    >>> E, P, Q = expmint.getEPQ2(A, .05, order=0, B=B)
    >>> E
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> P
    array([[ 0.0034],
           [ 0.0583],
           [ 0.0133]])
    >>> Q
    0.0
    """
    n = A.shape[0]
    if B is not None:
        i = B.shape[1]
    else:
        if half:
            if n & 1:
                raise ValueError(
                    "`A` must have an even number "
                    "of rows/cols (or use ``half=False``"
                )
            i = n // 2
        else:
            i = n
        B = np.eye(i)
    r = B.shape[0]
    Ah = A * h
    Bh = B * h
    if order == 1:
        N = n + 2 * i
        M = np.zeros((N, N), float)
        M[:n, :n] = Ah
        M[:r, n : n + i] = Bh
        M[n : n + i, n + i :] = np.eye(i)
        # start = time.time()
        # EM1 = la.expm(M, order)
        # print('1 la.expm et = ', time.time()-start)

        # start = time.time()
        EM = _expm_SS(M, Ah, order)
        # print('1 expm_SS et = ', time.time()-start)
        # print('error :', abs(EM-EM1).max())
        E = EM[:n, :n]
        Q = EM[:n, n + i :]
        P = EM[:n, n : n + i] - Q
    elif order == 0:
        M = np.zeros((n + i, n + i), float)
        M[:n, :n] = Ah
        M[:r, n:] = Bh
        # start = time.time()
        # EM1 = la.expm(M, order)
        # print('0 la.expm et = ', time.time()-start)

        # start = time.time()
        EM = _expm_SS(M, Ah, order)
        # print('0 expm_SS et = ', time.time()-start)
        # print('error :', abs(EM-EM1).max())
        E = EM[:n, :n]
        P = EM[:n, n:]
        Q = 0.0
    else:
        raise ValueError("`order` must be 0 or 1")
    return E, P, Q


def getEPQ(A, h, order=1, B=None, half=False):
    """
    Returns E, P, Q for the exponential solver given the state-space
    matrix `A`.

    Parameters
    ----------
    A : 2d ndarray
        The state-space matrix:  ``xdot = A x + B u``
    h : scalar
        Time step.
    order : integer, optional

        - 0 for the zero order hold (force stays constant across
          time step)
        - 1 for the 1st order hold (force can vary linearly across
          time step)

    B : d2 ndarray or None; optional
        If array, it multiplies the inputs; if None, it is assumed
        identity.
    half : bool; optional
        If `B` is a 2d ndarray, `half` is ignored. Otherwise, if
        `half` is False, a full size identity (same size as `A`) is
        used for `B`. If `half` is True, only the first half of the
        columns are retained (which is handy for converting a 2nd
        order ODE into a 1st order ODE as
        :class:`pyyeti.ode.SolveExp2` does -- where there are force
        inputs only for the first half of the equations).

    Returns
    -------
    E, P, Q : 2d ndarrays, except if ``order == 0``, ``Q = 0.``
        These are the coefficient matrices used to solve the ODE::

            for j in range(nt):
                d[:, j+1] = E*d[:, j] + P*F[:, j] + Q*F[:, j+1]

    Notes
    -----
    This routine calls either :func:`getEPQ1` or :func:`getEPQ2` for
    the bulk of the work. If the 1-norm of `A` is less than
    2.097847961257068 [#exp4]_, :func:`getEPQ1` is called; otherwise,
    :func:`getEPQ2` is called.

    References
    ----------
    .. [#exp4] Nicholas J. Higham (2005)
          "The Scaling and Squaring Method for the Matrix Exponential
          Revisited."
          SIAM Journal on Matrix Analysis and Applications.
          Vol 26, No. 4, pp 1179-1193.

    See also
    --------
    :func:`getEPQ_pow`, :class:`pyyeti.ode.SolveExp1`,
    :class:`pyyeti.ode.SolveExp2`

    Examples
    --------
    >>> from pyyeti import expmint
    >>> import numpy as np
    >>> import scipy.linalg as la
    >>> np.set_printoptions(4)
    >>> A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> B = np.array([[0, 1, 0]]).T
    >>> E, P, Q = expmint.getEPQ(A, .05, order=1, B=B)
    >>> E
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> P
    array([[ 0.0024],
           [ 0.0308],
           [ 0.0091]])
    >>> Q
    array([[ 0.0011],
           [ 0.0276],
           [ 0.0041]])
    >>> E, P, Q = expmint.getEPQ(A, .05, order=0, B=B)
    >>> E
    array([[ 1.0996,  0.1599,  0.2202],
           [ 0.3099,  1.3849,  0.46  ],
           [ 0.5202,  0.61  ,  1.6998]])
    >>> P
    array([[ 0.0034],
           [ 0.0583],
           [ 0.0133]])
    >>> Q
    0.0
    """
    norm1 = h * np.linalg.norm(A, 1)
    if norm1 <= 2.097847961257068:
        return getEPQ1(A, h, order, B, half)
    return getEPQ2(A, h, order, B, half)



# -*- coding: utf-8 -*-
import scipy.linalg as la
import numpy as np
from ._base_ode_class import _BaseODE


# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


class FreqDirect(_BaseODE):
    """
    2nd order ODE frequency domain solver

    This class is for solving::

        m xdd + b xd + k x = f

    Notes
    -----
    Each frequency is solved via complex matrix inversion. There
    is no partitioning for rigid-body modes or residual-
    flexibility modes. Note that the solution will be fast if all
    matrices are diagonal.

    Unlike :class:`SolveUnc`, this routine makes no special provisions
    for rigid-body modes when computing the response; therefore,
    including 0.0 in `freq` can cause a divide-by-zero. It is
    therefore recommended to ensure that all values in `freq` > 0.0,
    at least when rigid-body modes are present. After the solution is
    computed, for equations that are in modal space, the rigid-body
    part of the response may be zeroed out according to the `incrb`
    parameter in :func:`fsolve`.

    See also
    --------
    :func:`SolveUnc.fsolve`

    Examples
    --------
    .. plot::
        :context: close-figs

        >>> from pyyeti import ode
        >>> import numpy as np
        >>> m = np.array([10., 30., 30., 30.])    # diag mass
        >>> k = np.array([0., 6.e5, 6.e5, 6.e5])  # diag stiffness
        >>> zeta = np.array([0., .05, 1., 2.])    # % damping
        >>> b = 2.*zeta*np.sqrt(k/m)*m            # diag damping
        >>> freq = np.arange(.1, 35, .1)          # frequency
        >>> f = 100*np.ones((4, freq.size))       # constant ffn
        >>> ts = ode.FreqDirect(m, b, k)
        >>> sol = ts.fsolve(f, freq)

        Solve @ 25 Hz by hand for comparison:

        >>> w = 2*np.pi*25
        >>> i = np.argmin(abs(freq-25))
        >>> H = -w**2*m + 1j*w*b + k
        >>> disp = f[:, i] / H
        >>> velo = 1j*w*disp
        >>> acce = -w**2*disp
        >>> np.allclose(disp, sol.d[:, i])
        True
        >>> np.allclose(velo, sol.v[:, i])
        True
        >>> np.allclose(acce, sol.a[:, i])
        True

        Plot the four accelerations:

        >>> import matplotlib.pyplot as plt
        >>> fig = plt.figure('Example', figsize=[8, 8])
        >>> fig.clf()
        >>> labels = ['Rigid-body', 'Underdamped',
        ...           'Critically Damped', 'Overdamped']
        >>> for j, name in zip(range(4), labels):
        ...     _ = plt.subplot(4, 1, j+1)
        ...     _ = plt.plot(freq, abs(sol.a[j]))
        ...     _ = plt.title(name)
        ...     _ = plt.ylabel('Acceleration')
        ...     _ = plt.xlabel('Frequency (Hz)')
        >>> fig.tight_layout()
    """

    def __init__(self, m, b, k, rb=None, rf=None):
        """
        Instantiates a :class:`FreqDirect` solver.

        Parameters
        ----------
        m : 1d or 2d ndarray or None
            Mass; vector (of diagonal), or full; if None, mass is
            assumed identity
        b : 1d or 2d ndarray
            Damping; vector (of diagonal), or full
        k : 1d or 2d ndarray
            Stiffness; vector (of diagonal), or full
        rb : 1d array or None; optional
            An option for equations in modal space. Index or bool
            partition vector for rigid-body modes. Set to [] to
            specify no rigid-body modes. If None, the rigid-body modes
            will be automatically detected by this logic for uncoupled
            systems::

               rb = np.nonzero(abs(k).max(0) < 0.005)[0]

            And by this logic for coupled systems::

               rb = ((abs(k).max(axis=0) < 0.005) &
                     (abs(k).max(axis=1) < 0.005) &
                     (abs(b).max(axis=0) < 0.005) &
                     (abs(b).max(axis=1) < 0.005)).nonzero()[0]

        rf : 1d array or None; optional
            Index or bool partition vector for residual-flexibility
            modes; these will be solved statically. As for the `rb`
            option, the `rf` option only applies to modal space
            equations.

        Notes
        -----
        The instance is populated with some or all of the following
        members.

        ========   ==================================================
         Member    Description
        ========   ==================================================
         m         mass
         b         damping
         k         stiffness
         h         time step (None)
         rb        index vector or slice for the rb modes
         el        index vector or slice for the el modes
         rf        index vector or slice for the rf modes ([])
         _rb       index vector or slice for the rb modes relative to
                   the non-rf modes
         _el       index vector or slice for the el modes relative to
                   the non-rf modes
         nonrf     index vector or slice for the non-rf modes
         kdof      index vector or slice for the non-rf modes
         n         number of total DOF
         rfsize    number of rf modes
         nonrfsz   number of non-rf modes
         ksize     number of non-rf modes
         krf       stiffness for the rf modes
         ikrf      inverse of stiffness for the rf modes
         unc       True if there are no off-diagonal terms in any
                   matrix; False otherwise
         systype   float or complex; determined by `m`, `b`, and `k`
        ========   ==================================================

        The mass, damping and stiffness may be real or complex. This
        routine currently does not accept the `rf` input (if there are
        any, they are treated like all other elastic modes).
        """
        self._common_precalcs(m, b, k, h=None, rb=rb, rf=rf)
        self._mk_slices()  # dorbel=False)

    def fsolve(self, force, freq, incrb=2, rf_disp_only=False):
        """
        Solve equations of motion in frequency domain.

        Parameters
        ----------
        force : 2d ndarray
            The force matrix; ndof x freq
        freq : 1d ndarray
            Frequency vector in Hz; solution will be computed at all
            frequencies in `freq`
        incrb : 0, 1, or 2; optional
            Specifies how to handle rigid-body responses:

            ======  ==============================================
            incrb   description
            ======  ==============================================
               0    no rigid-body is included
               1    acceleration and velocity rigid-body only
               2    all of rigid-body is included
            ======  ==============================================

        rf_disp_only : bool; optional
            This option specifies how to handle the velocity and
            acceleration terms for residual-flexibility modes. If
            True, they are set to zero. If False, they are computed
            from the normal frequency-domain relationships::

                velocity = i * omega * displacement
                acceleration = -omega ** 2 * displacement

        Returns
        -------
        A SimpleNamespace with the members:

        d : 2d ndarray
            Displacement; ndof x freq
        v : 2d ndarray
            Velocity; ndof x freq
        a : 2d ndarray
            Acceleration; ndof x freq
        f : 1d ndarray
            Frequency vector (same as the input `freq`)

        Notes
        -----
        See :class:`FreqDirect` for more discussion on how rigid-body
        response is handled.
        """
        force = np.atleast_2d(force)
        freq = np.atleast_1d(freq)
        d, v, a, force = self._init_dva(
            force,
            d0=None,
            v0=None,
            static_ic=False,
            istime=False,
            freq=freq,
            rf_disp_only=rf_disp_only,
        )

        if self.ksize == 0:
            return self._solution_freq(d, v, a, freq)

        self._force_freq_compat_chk(force, freq)
        m, b, k = self.m, self.b, self.k
        kdof = self.kdof
        force = force[kdof]
        if self.unc:
            # equations are uncoupled, solve everything in one step:
            Omega = 2 * np.pi * freq[None, :]
            if m is None:
                H = ((1j * b)[:, None] @ Omega + k[:, None]) - Omega ** 2
            else:
                H = (1j * b)[:, None] @ Omega + k[:, None] - m[:, None] @ Omega ** 2
            d[kdof] = force / H
        else:
            # equations are coupled, use a loop:
            Omega = 2 * np.pi * freq
            if m is None:
                m = np.eye(self.ksize)
            for i, O in enumerate(Omega):
                Hi = 1j * b * O + k - m * O ** 2
                d[kdof, i] = la.solve(Hi, force[:, i])
        a[kdof] = -(Omega ** 2) * d[kdof]
        v[kdof] = 1j * Omega * d[kdof]

        if incrb < 2:
            d[self.rb] = 0
            if incrb == 0:
                a[self.rb] = 0
                v[self.rb] = 0
        return self._solution_freq(d, v, a, freq)
